import gzip
import json
import logging
import os
import re
import subprocess
import sys
import threading
from typing import Dict, Optional

from requests.models import Response

from localstack import config
from localstack.constants import (
    HEADER_LOCALSTACK_EDGE_URL,
    HEADER_LOCALSTACK_REQUEST_URL,
    INTERNAL_AWS_ACCESS_KEY_ID,
    LOCALHOST,
    LOCALHOST_IP,
    LOCALSTACK_ROOT_FOLDER,
    LS_LOG_TRACE_INTERNAL,
    PATH_USER_REQUEST,
)
from localstack.http import Router
from localstack.http.dispatcher import Handler, handler_dispatcher
from localstack.services.generic_proxy import ProxyListener, modify_and_forward, start_proxy_server
from localstack.services.infra import PROXY_LISTENERS
from localstack.services.plugins import SERVICE_PLUGINS
from localstack.services.s3.s3_utils import uses_host_addressing
from localstack.services.sqs.sqs_listener import is_sqs_queue_url
from localstack.utils import common, persistence
from localstack.utils.aws import aws_stack
from localstack.utils.aws.aws_stack import is_internal_call_context, set_default_region_in_headers
from localstack.utils.aws.request_routing import extract_version_and_action, matches_service_action
from localstack.utils.common import (
    TMP_THREADS,
    empty_context_manager,
    get_service_protocol,
    is_port_open,
    is_root,
    parse_request_data,
    run,
)
from localstack.utils.common import safe_requests as requests
from localstack.utils.common import sleep_forever, start_thread, to_bytes, to_str, truncate
from localstack.utils.server.http2_server import HTTPErrorResponse

LOG = logging.getLogger(__name__)

# Header to indicate that the process should kill itself. This is required because if
# this process is started as root, then we cannot kill it from a non-root process
HEADER_KILL_SIGNAL = "x-localstack-kill"

# Header to indicate the current API (service) being called
HEADER_TARGET_API = "x-localstack-tgt-api"

# lock obtained during boostrapping (persistence restoration) to avoid concurrency issues
BOOTSTRAP_LOCK = threading.RLock()

PORT_DNS = 53

GZIP_ENCODING = "GZIP"
IDENTITY_ENCODING = "IDENTITY"
S3 = "s3"
API_UNKNOWN = "_unknown_"
# APIs for which no gzip encoding should be applied when returning the response
HEADER_SKIP_RESPONSE_ZIPPING = "_skip_response_gzipping_"
SKIP_GZIP_APIS = [S3]


class ProxyListenerEdge(ProxyListener):
    def __init__(self, service_manager=None) -> None:
        super().__init__()
        self.service_manager = service_manager or SERVICE_PLUGINS

    def forward_request(self, method, path, data, headers):

        if common.INFRA_STOPPED:
            return 503

        if config.EDGE_FORWARD_URL:
            return do_forward_request_network(
                0, method, path, data, headers, target_url=config.EDGE_FORWARD_URL
            )

        # kill the process if we receive this header
        headers.get(HEADER_KILL_SIGNAL) and sys.exit(0)

        target = headers.get("x-amz-target", "")
        auth_header = get_auth_string(method, path, headers, data)
        if auth_header and not headers.get("authorization"):
            headers["authorization"] = auth_header
        host = headers.get("host", "")
        orig_req_url = headers.pop(HEADER_LOCALSTACK_REQUEST_URL, "")
        headers[HEADER_LOCALSTACK_EDGE_URL] = (
            re.sub(r"^([^:]+://[^/]+).*", r"\1", orig_req_url) or "http://%s" % host
        )

        # extract API details
        api, port, path, host = get_api_from_headers(headers, method=method, path=path, data=data)

        set_default_region_in_headers(headers)

        if port and int(port) < 0:
            return 404

        if not port:
            api, port = get_api_from_custom_rules(method, path, data, headers) or (
                api,
                port,
            )

        should_log_trace = is_trace_logging_enabled(headers)
        if api and should_log_trace:
            # print request trace for debugging, if enabled
            LOG.debug(
                'IN(%s): "%s %s" - headers: %s - data: %s', api, method, path, dict(headers), data
            )

        if not port:
            if method == "OPTIONS":
                if api and should_log_trace:
                    # print request trace for debugging, if enabled
                    LOG.debug('IN(%s): "%s %s" - status: %s', api, method, path, 200)
                return 200

            if api in ["", None, API_UNKNOWN]:
                truncated = truncate(data)
                if auth_header or target or data or path not in ["/", "/favicon.ico"]:
                    LOG.info(
                        (
                            'Unable to find forwarding rule for host "%s", path "%s %s", '
                            'target header "%s", auth header "%s", data "%s"'
                        ),
                        host,
                        method,
                        path,
                        target,
                        auth_header,
                        truncated,
                    )
            else:
                LOG.info(
                    (
                        'Unable to determine forwarding port for API "%s" - please '
                        "make sure this API is enabled via the SERVICES configuration"
                    ),
                    api,
                )
            response = Response()
            response.status_code = 404
            response._content = '{"status": "running"}'
            return response

        if api and not headers.get("Authorization"):
            headers["Authorization"] = aws_stack.mock_aws_request_headers(api)["Authorization"]
        headers[HEADER_TARGET_API] = str(api)

        headers["Host"] = host
        if isinstance(data, dict):
            data = json.dumps(data)

        encoding_type = headers.get("Content-Encoding") or ""
        if encoding_type.upper() == GZIP_ENCODING.upper() and api not in SKIP_GZIP_APIS:
            headers.set("Content-Encoding", IDENTITY_ENCODING)
            data = gzip.decompress(data)

        is_internal_call = is_internal_call_context(headers)

        self._require_service(api)

        lock_ctx = BOOTSTRAP_LOCK
        if is_internal_call or persistence.is_persistence_restored():
            lock_ctx = empty_context_manager()

        with lock_ctx:
            result = do_forward_request(api, method, path, data, headers, port=port)
            if should_log_trace and result not in [None, False, True]:
                result_status_code = getattr(result, "status_code", result)
                result_headers = getattr(result, "headers", {})
                result_content = getattr(result, "content", "")
                LOG.debug(
                    'OUT(%s): "%s %s" - status: %s - response headers: %s - response: %s',
                    api,
                    method,
                    path,
                    result_status_code,
                    dict(result_headers or {}),
                    result_content,
                )
            return result

    def return_response(self, method, path, data, headers, response):
        api = headers.get(HEADER_TARGET_API) or ""

        if is_trace_logging_enabled(headers):
            # print response trace for debugging, if enabled
            if api and api != API_UNKNOWN:
                LOG.debug(
                    'OUT(%s): "%s %s" - status: %s - response headers: %s - response: %s',
                    api,
                    method,
                    path,
                    response.status_code,
                    dict(response.headers),
                    response.content,
                )

        if (
            response._content
            and headers.get("Accept-Encoding") == "gzip"
            and api not in SKIP_GZIP_APIS
            and not response.headers.pop(HEADER_SKIP_RESPONSE_ZIPPING, None)
        ):
            # services may decide to set HEADER_SKIP_RESPONSE_ZIPPING in the response, to skip result transformations
            response._content = gzip.compress(to_bytes(response._content))
            response.headers["Content-Length"] = str(len(response._content))
            response.headers["Content-Encoding"] = "gzip"

    def _require_service(self, api):
        if not self.service_manager.exists(api):
            raise HTTPErrorResponse("no provider exists for service %s" % api, code=500)

        try:
            self.service_manager.require(api)
        except Exception as e:
            raise HTTPErrorResponse("failed to get service for %s: %s" % (api, e), code=500)


def do_forward_request(api, method, path, data, headers, port=None):
    if config.FORWARD_EDGE_INMEM:
        result = do_forward_request_inmem(api, method, path, data, headers, port=port)
    else:
        result = do_forward_request_network(port, method, path, data, headers)
    if hasattr(result, "status_code") and int(result.status_code) >= 400 and method == "OPTIONS":
        # fall back to successful response for OPTIONS requests
        return 200
    return result


def get_handler_for_api(api, headers):
    return PROXY_LISTENERS.get(api)


def do_forward_request_inmem(api, method, path, data, headers, port=None):
    listener_details = get_handler_for_api(api, headers)
    if not listener_details:
        message = f'Unable to find listener for service "{api}" - please make sure to include it in $SERVICES'
        LOG.warning(message)
        raise HTTPErrorResponse(message, code=400)
    service_name, backend_port, listener = listener_details
    # TODO determine client address..?
    client_address = LOCALHOST_IP
    server_address = headers.get("host") or LOCALHOST
    forward_url = "http://%s:%s" % (LOCALHOST, backend_port)
    response = modify_and_forward(
        method=method,
        path=path,
        data_bytes=data,
        headers=headers,
        forward_base_url=forward_url,
        listeners=[listener],
        client_address=client_address,
        server_address=server_address,
    )
    return response


def do_forward_request_network(port, method, path, data, headers, target_url=None):
    # TODO: enable per-service endpoints, to allow deploying in distributed settings
    target_url = target_url or "%s://%s:%s" % (get_service_protocol(), LOCALHOST, port)
    url = "%s%s" % (target_url, path)
    response = requests.request(
        method, url, data=data, headers=headers, verify=False, stream=True, allow_redirects=False
    )
    return response


def get_auth_string(method, path, headers, data=None):
    """
    Get Auth header from Header (this is how aws client's like boto typically
    provide it) or from query string or url encoded parameters (sometimes
    happens with presigned requests. Always return in the Authorization Header
    form.

    Typically an auth string comes in as a header:

        Authorization: AWS4-HMAC-SHA256 \
        Credential=_not_needed_locally_/20210312/us-east-1/sqs/aws4_request, \
        SignedHeaders=content-type;host;x-amz-date, \
        Signature=9277c941f4ecafcc0f290728e50cd7a3fa0e41763fbd2373fcdd3faf2dbddc2e

    Here's what Authorization looks like as part of an presigned GET request:

       &X-Amz-Algorithm=AWS4-HMAC-SHA256\
       &X-Amz-Credential=test%2F20210313%2Fus-east-1%2Fsqs%2Faws4_request\
       &X-Amz-Date=20210313T011059Z&X-Amz-Expires=86400000&X-Amz-SignedHeaders=host\
       &X-Amz-Signature=2c652c7bc9a3b75579db3d987d1e6dd056f0ac776c1e1d4ec91e2ce84e5ad3ae
    """

    auth_header = headers.get("authorization", "")

    if auth_header:
        return auth_header

    data_components = parse_request_data(method, path, data)
    algorithm = data_components.get("X-Amz-Algorithm")
    credential = data_components.get("X-Amz-Credential")
    signature = data_components.get("X-Amz-Signature")
    signed_headers = data_components.get("X-Amz-SignedHeaders")

    if algorithm and credential and signature and signed_headers:
        return (
            f"{algorithm} Credential={credential}, "
            + f"SignedHeaders={signed_headers}, "
            + f"Signature={signature}"
        )

    return ""


# TODO: refactor this function -> returning the port is redundant (given the returned service name)
def get_api_from_headers(headers, method=None, path=None, data=None):
    """Determine API and backend port based on "Authorization" or "Host" headers."""

    # initialize result
    result = API_UNKNOWN, 0

    target = headers.get("x-amz-target", "")
    host = headers.get("host", "")
    auth_header = headers.get("authorization", "")

    if not auth_header and not target and "." not in host:
        return result[0], result[1], path, host

    path = path or "/"

    # https://docs.aws.amazon.com/general/latest/gr/sigv4-signed-request-examples.html
    try:
        service = extract_service_name_from_auth_header(headers)
        assert service
        result = service, get_service_port_for_account(service, headers)
    except Exception:
        pass

    result_before = result

    # Fallback rules and route customizations applied below
    if host.endswith("cloudfront.net"):
        path = path or "/"
        result = "cloudfront", config.service_port("cloudfront")
    elif target.startswith("AWSCognitoIdentityProviderService") or "cognito-idp." in host:
        result = "cognito-idp", config.service_port("cognito-idp")
    elif target.startswith("AWSCognitoIdentityService") or "cognito-identity." in host:
        result = "cognito-identity", config.service_port("cognito-identity")
    elif result[0] == "s3" or uses_host_addressing(headers):
        result = "s3", config.service_port("s3")
    elif result[0] == "states" in auth_header or host.startswith("states."):
        result = "stepfunctions", config.service_port("stepfunctions")
    elif "route53." in host:
        result = "route53", config.service_port("route53")
    elif result[0] == "monitoring":
        result = "cloudwatch", config.service_port("cloudwatch")
    elif result[0] == "email":
        result = "ses", config.service_port("ses")
    elif result[0] == "execute-api" or ".execute-api." in host:
        result = "apigateway", config.service_port("apigateway")
    elif target.startswith("Firehose_"):
        result = "firehose", config.service_port("firehose")
    elif target.startswith("DynamoDB_"):
        result = "dynamodb", config.service_port("dynamodb")
    elif target.startswith("DynamoDBStreams") or host.startswith("streams.dynamodb."):
        # Note: DDB streams requests use ../dynamodb/.. auth header, hence we also need to update result_before
        result = result_before = "dynamodbstreams", config.service_port("dynamodbstreams")
    elif result[0] == "EventBridge" or target.startswith("AWSEvents"):
        result = "events", config.service_port("events")
    elif target.startswith("ResourceGroupsTaggingAPI_"):
        result = "resourcegroupstaggingapi", config.service_port("resourcegroupstaggingapi")
    elif result[0] == "resource-groups":
        result = "resource-groups", config.service_port("resource-groups")
    elif result[0] == "es" and path is not None and not path.startswith("/2015-01-01/"):
        # For OpenSearch, the auth header points to the API ("es").
        # However, if the path does _not_ start with /2015-01-01 (the API version path prefix for the only ES API
        # version) it is a request to the opensearch API.
        result = "opensearch", config.service_port("opensearch")

    return result[0], result_before[1] or result[1], path, host


def extract_service_name_from_auth_header(headers: Dict) -> Optional[str]:
    try:
        auth_header = headers.get("authorization", "")
        credential_scope = auth_header.split(",")[0].split()[1]
        _, _, _, service, _ = credential_scope.split("/")
        return service
    except Exception:
        return


def is_s3_form_data(data_bytes):
    if to_bytes("key=") in data_bytes:
        return True
    if (
        to_bytes("Content-Disposition: form-data") in data_bytes
        and to_bytes('name="key"') in data_bytes
    ):
        return True
    return False


# TODO: refactor this function -> returning the port is redundant (given the returned service name)
def get_api_from_custom_rules(method, path, data, headers):
    """Determine backend port based on custom rules."""

    # API Gateway invocation URLs
    if ("/%s/" % PATH_USER_REQUEST) in path:
        return "apigateway", config.service_port("apigateway")

    # detect S3 presigned URLs
    if "AWSAccessKeyId=" in path or "Signature=" in path:
        return "s3", config.service_port("s3")

    # heuristic for SQS queue URLs
    if is_sqs_queue_url(path):
        return "sqs", config.service_port("sqs")

    # DynamoDB shell URLs
    if path.startswith("/shell") or path.startswith("/dynamodb/shell"):
        return "dynamodb", config.service_port("dynamodb")

    data_bytes = to_bytes(data or "")
    version, action = extract_version_and_action(path, data_bytes)

    def _in_path_or_payload(search_str):
        return to_str(search_str) in path or to_bytes(search_str) in data_bytes

    if path == "/" and b"QueueName=" in data_bytes:
        return "sqs", config.service_port("sqs")

    if "Action=ConfirmSubscription" in path:
        return "sns", config.service_port("sns")

    if path.startswith("/2015-03-31/functions/"):
        return "lambda", config.service_port("lambda")

    if _in_path_or_payload("Action=AssumeRoleWithWebIdentity"):
        return "sts", config.service_port("sts")

    if _in_path_or_payload("Action=AssumeRoleWithSAML"):
        return "sts", config.service_port("sts")

    if _in_path_or_payload("Action=AssumeRole"):
        return "sts", config.service_port("sts")

    # SQS queue requests
    if _in_path_or_payload("QueueUrl=") and _in_path_or_payload("Action="):
        return "sqs", config.service_port("sqs")
    if matches_service_action("sqs", action, version=version):
        return "sqs", config.service_port("sqs")

    # SNS topic requests
    if matches_service_action("sns", action, version=version):
        return "sns", config.service_port("sns")

    # TODO: move S3 public URLs to a separate port/endpoint, OR check ACLs here first
    stripped = path.strip("/")
    if method in ["GET", "HEAD"] and stripped:
        # assume that this is an S3 GET request with URL path `/<bucket>/<key ...>`
        return "s3", config.service_port("s3")

    # detect S3 URLs
    if stripped and "/" not in stripped:
        if method == "PUT":
            # assume that this is an S3 PUT bucket request with URL path `/<bucket>`
            return "s3", config.service_port("s3")
        if method == "POST" and is_s3_form_data(data_bytes):
            # assume that this is an S3 POST request with form parameters or multipart form in the body
            return "s3", config.service_port("s3")

    # detect S3 requests sent from aws-cli using --no-sign-request option
    if "aws-cli/" in headers.get("User-Agent", ""):
        return "s3", config.service_port("s3")

    # S3 delete object requests
    if (
        method == "POST"
        and "delete=" in path
        and b"<Delete" in data_bytes
        and b"<Key>" in data_bytes
    ):
        return "s3", config.service_port("s3")

    # Put Object API can have multiple keys
    if stripped.count("/") >= 1 and method == "PUT":
        # assume that this is an S3 PUT bucket object request with URL path `/<bucket>/object`
        # or `/<bucket>/object/object1/+`
        return "s3", config.service_port("s3")

    auth_header = headers.get("Authorization") or ""

    # detect S3 requests with "AWS id:key" Auth headers
    if auth_header.startswith("AWS "):
        return "s3", config.service_port("s3")

    # certain EC2 requests from Java SDK contain no Auth headers (issue #3805)
    if b"Version=2016-11-15" in data_bytes:
        return "ec2", config.service_port("ec2")


def get_service_port_for_account(service, headers):
    # assume we're only using a single account, hence return the static port mapping from config.py
    return config.service_port(service)


PROXY_LISTENER_EDGE = ProxyListenerEdge()
# the ROUTER is part of the edge proxy. use the router to inject custom handlers that are handled before actual
# service calls
ROUTER: Router[Handler] = Router(dispatcher=handler_dispatcher())


def is_trace_logging_enabled(headers):
    if not config.LS_LOG:
        return False
    if config.LS_LOG == LS_LOG_TRACE_INTERNAL:
        return True
    auth_header = headers.get("Authorization") or ""
    return INTERNAL_AWS_ACCESS_KEY_ID not in auth_header


def do_start_edge(bind_address, port, use_ssl, asynchronous=False):
    from localstack.http.adapters import RouterListener
    from localstack.services.internal import LocalstackResourceHandler

    start_dns_server(asynchronous=True)

    listeners = [
        LocalstackResourceHandler(),  # handle internal resources first
        RouterListener(ROUTER),  # then custom routes
        PROXY_LISTENER_EDGE,  # then call the edge proxy listener
    ]

    # get port and start Edge
    print("Starting edge router (http%s port %s)..." % ("s" if use_ssl else "", port))
    # use use_ssl=True here because our proxy allows both, HTTP and HTTPS traffic
    proxy = start_proxy_server(
        port,
        bind_address=bind_address,
        use_ssl=True,
        update_listener=listeners,
        check_port=False,
    )
    if not asynchronous:
        proxy.join()
    return proxy


def can_use_sudo():
    try:
        run("sudo -n -v", print_error=False)
        return True
    except Exception:
        return False


def ensure_can_use_sudo():
    if not is_root() and not can_use_sudo():
        if not sys.stdin.isatty():
            raise IOError("cannot get sudo password from non-tty input")
        print("Please enter your sudo password (required to configure local network):")
        run("sudo -v", stdin=True)


def start_component(component: str, port=None, use_ssl=True, asynchronous=False):
    if component == "edge":
        return start_edge(port=port, use_ssl=use_ssl, asynchronous=asynchronous)
    if component == "dns":
        return start_dns_server(asynchronous=asynchronous)
    raise Exception("Unexpected component name '%s' received during start up" % component)


def start_dns_server(asynchronous=False):
    try:
        # start local DNS server, if present
        from localstack_ext import config as config_ext
        from localstack_ext.services import dns_server

        if config_ext.DNS_ADDRESS in config.FALSE_STRINGS:
            return

        if is_port_open(PORT_DNS):
            return

        if is_root():
            result = dns_server.start_servers()
            if not asynchronous:
                sleep_forever()
            return result

        env_vars = {}
        for env_var in config.CONFIG_ENV_VARS:
            if env_var.startswith("DNS_"):
                value = os.environ.get(env_var, None)
                if value is not None:
                    env_vars[env_var] = value

        # note: running in a separate process breaks integration with Route53 (to be fixed for local dev mode!)
        return run_process_as_sudo("dns", PORT_DNS, asynchronous=asynchronous, env_vars=env_vars)
    except Exception:
        pass


def start_edge(port=None, use_ssl=True, asynchronous=False):
    if not port:
        port = config.EDGE_PORT
    if config.EDGE_PORT_HTTP and config.EDGE_PORT_HTTP != port:
        do_start_edge(
            config.EDGE_BIND_HOST,
            config.EDGE_PORT_HTTP,
            use_ssl=False,
            asynchronous=True,
        )
    if port > 1024 or is_root():
        return do_start_edge(config.EDGE_BIND_HOST, port, use_ssl, asynchronous=asynchronous)

    # process requires privileged port but we're not root -> try running as sudo

    class Terminator(object):
        def stop(self, quiet=True):
            try:
                url = "http%s://%s:%s" % ("s" if use_ssl else "", LOCALHOST, port)
                requests.verify_ssl = False
                requests.post(url, headers={HEADER_KILL_SIGNAL: "kill"})
            except Exception:
                pass

    # register a signal handler to terminate the sudo process later on
    TMP_THREADS.append(Terminator())

    return run_process_as_sudo("edge", port, asynchronous=asynchronous)


def run_process_as_sudo(component, port, asynchronous=False, env_vars=None):
    # make sure we can run sudo commands
    try:
        ensure_can_use_sudo()
    except Exception as e:
        LOG.error("cannot start service on privileged port %s: %s", port, str(e))
        return

    # prepare environment
    env_vars = env_vars or {}
    env_vars["PYTHONPATH"] = f".:{LOCALSTACK_ROOT_FOLDER}"
    env_vars["EDGE_FORWARD_URL"] = config.get_edge_url()
    env_vars["EDGE_BIND_HOST"] = config.EDGE_BIND_HOST
    env_vars_str = env_vars_to_string(env_vars)

    # start the process as sudo
    sudo_cmd = "sudo -n"
    python_cmd = sys.executable
    cmd = [
        sudo_cmd,
        env_vars_str,
        python_cmd,
        "-m",
        "localstack.services.edge",
        component,
        str(port),
    ]
    shell_cmd = " ".join(cmd)

    def run_command(*_):
        run(shell_cmd, outfile=subprocess.PIPE, print_error=False, env_vars=env_vars)

    LOG.debug("Running command as sudo: %s", shell_cmd)
    result = start_thread(run_command, quiet=True) if asynchronous else run_command()
    return result


def env_vars_to_string(env_vars: Dict) -> str:
    return " ".join(f"{k}='{v}'" for k, v in env_vars.items())


if __name__ == "__main__":
    logging.basicConfig()
    start_component(sys.argv[1], int(sys.argv[2]))



import logging
import re
import threading
from urllib.parse import urlparse

from flask import request
from requests.models import Request
from requests.structures import CaseInsensitiveDict

from localstack import config
from localstack.constants import APPLICATION_JSON, APPLICATION_XML, HEADER_CONTENT_TYPE
from localstack.utils.aws import aws_stack
from localstack.utils.aws.aws_responses import (
    is_json_request,
    requests_error_response,
    requests_response,
    requests_to_flask_response,
)
from localstack.utils.common import snake_to_camel_case
from localstack.utils.patch import patch
from localstack.utils.run import FuncThread

LOG = logging.getLogger(__name__)

THREAD_LOCAL = threading.local()

MARKER_APIGW_REQUEST_REGION = "__apigw_request_region__"


def get_proxy_request_for_thread():
    try:
        return THREAD_LOCAL.request_context
    except Exception:
        return None


def get_flask_request_for_thread():
    try:
        # Append/cache a converted request (requests.Request) to the the thread-local Flask request.
        #  We use this request object as the invocation context, which may be modified in other places,
        #  e.g., when manually configuring the region in the request context of an incoming API call.
        if not hasattr(request, "_converted_request"):
            request._converted_request = Request(
                url=request.path,
                data=request.data,
                headers=CaseInsensitiveDict(request.headers),
                method=request.method,
            )
        return request._converted_request
    except Exception as e:
        # swallow error: "Working outside of request context."
        if "Working outside" in str(e):
            return None
        raise


def extract_region_from_auth_header(headers):
    auth = headers.get("Authorization") or ""
    region = re.sub(r".*Credential=[^/]+/[^/]+/([^/]+)/.*", r"\1", auth)
    if region == auth:
        return None
    return region


def extract_region_from_headers(headers):
    region = headers.get(MARKER_APIGW_REQUEST_REGION)
    # Fix region lookup for certain requests, e.g., API gateway invocations
    #  that do not contain region details in the Authorization header.

    if region:
        return region

    region = extract_region_from_auth_header(headers)

    if not region:
        # fall back to local region
        region = aws_stack.get_local_region()

    return region


def get_request_context():
    candidates = [get_proxy_request_for_thread(), get_flask_request_for_thread()]
    for req in candidates:
        if req is not None:
            return req


class RequestContextManager(object):
    """Context manager which sets the given request context (i.e., region) for the scope of the block."""

    def __init__(self, request_context):
        self.request_context = request_context

    def __enter__(self):
        THREAD_LOCAL.request_context = self.request_context

    def __exit__(self, type, value, traceback):
        THREAD_LOCAL.request_context = None


def get_region_from_request_context():
    """look up region from request context"""

    if config.USE_SINGLE_REGION:
        return

    request_context = get_request_context()
    if not request_context:
        return

    return extract_region_from_headers(request_context.headers)


def configure_region_for_current_request(region_name: str, service_name: str):
    """Manually configure (potentially overwrite) the region in the current request context. This may be
    used by API endpoints that are invoked directly by the user (without specifying AWS Authorization
    headers), to still enable transparent region lookup via aws_stack.get_region() ..."""

    # TODO: leaving import here for now, to avoid circular dependency
    from localstack.utils.aws import aws_stack

    request_context = get_request_context()
    if not request_context:
        LOG.info(
            "Unable to set region '%s' in undefined request context: %s",
            region_name,
            request_context,
        )
        return

    headers = request_context.headers
    auth_header = headers.get("Authorization")
    auth_header = auth_header or aws_stack.mock_aws_request_headers(service_name)["Authorization"]
    auth_header = auth_header.replace("/%s/" % aws_stack.get_region(), "/%s/" % region_name)
    try:
        headers["Authorization"] = auth_header
    except Exception as e:
        if "immutable" not in str(e):
            raise
        _context_to_update = get_proxy_request_for_thread() or request
        _context_to_update.headers = CaseInsensitiveDict({**headers, "Authorization": auth_header})


def mock_request_for_region(region_name: str, service_name: str = "dummy") -> Request:
    result = Request()
    result.headers["Authorization"] = aws_stack.mock_aws_request_headers(
        service_name, region_name=region_name
    )["Authorization"]
    return result


def patch_moto_request_handling():
    # leave here to avoid import issues
    from moto.core import utils as moto_utils

    from localstack.services.edge import extract_service_name_from_auth_header

    # make sure we properly handle/propagate "not implemented" errors
    @patch(moto_utils.convert_to_flask_response.__call__)
    def convert_to_flask_response_call(fn, *args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except NotImplementedError as e:
            action = request.headers.get("X-Amz-Target")
            action = action or f"{request.method} {urlparse(request.url).path}"
            if action == "POST /":
                # try to extract action from exception string
                match = re.match(r"The ([a-zA-Z0-9_-]+) action has not been implemented", str(e))
                if match:
                    action = snake_to_camel_case(match.group(1))
            service = extract_service_name_from_auth_header(request.headers)
            msg = f"API action '{action}' for service '{service}' not yet implemented"
            response = requests_error_response(request.headers, msg, code=501)
            if config.MOCK_UNIMPLEMENTED:
                is_json = is_json_request(request.headers)
                headers = {HEADER_CONTENT_TYPE: APPLICATION_JSON if is_json else APPLICATION_XML}
                content = "{}" if is_json else "<Response />"  # TODO: return proper mocked response
                response = requests_response(content, headers=headers)
                LOG.info(f"{msg}. Returning mocked response due to MOCK_UNIMPLEMENTED=1")
            else:
                LOG.info(msg)
            # TODO: publish analytics event ...
            return requests_to_flask_response(response)

    if config.USE_SINGLE_REGION:
        return

    # make sure that we inherit THREAD_LOCAL request contexts to spawned sub-threads
    @patch(FuncThread.__init__)
    def thread_init(fn, self, *args, **kwargs):
        self._req_context = get_request_context()
        return fn(self, *args, **kwargs)

    @patch(FuncThread.run)
    def thread_run(fn, self, *args, **kwargs):
        if self._req_context:
            THREAD_LOCAL.request_context = self._req_context
        return fn(self, *args, **kwargs)



import types

from moto.acm import models as acm_models

from localstack import config
from localstack.services.infra import start_moto_server


def apply_patches():
    def describe(self):
        result = describe_orig(self)
        cert = result.get("Certificate", {})
        sans = cert.get("SubjectAlternativeNames", [])

        # add missing attributes in ACM certs that cause Terraform to fail
        addenda = {
            "RenewalEligibility": "INELIGIBLE",
            "KeyUsages": [{"Name": "DIGITAL_SIGNATURE"}],
            "ExtendedKeyUsages": [],
            "Options": {"CertificateTransparencyLoggingPreference": "ENABLED"},
        }
        addenda["DomainValidationOptions"] = options = (
            getattr(self, "domain_validation_options", None) or []
        )
        if not options:
            options = addenda["DomainValidationOptions"] = [
                {"ValidationMethod": cert.get("ValidationMethod")}
            ]
        for san in sans:
            if san != cert.get("DomainName"):
                options.append(
                    {
                        "DomainName": san,
                        "ValidationMethod": cert.get("ValidationMethod"),
                    }
                )
        for option in options:
            option["DomainName"] = domain_name = option.get("DomainName") or cert.get("DomainName")
            option["ValidationDomain"] = option.get("ValidationDomain") or option["DomainName"]
            option["ValidationMethod"] = option.get("ValidationMethod") or "DNS"
            status = option.get("ValidationStatus")
            option["ValidationStatus"] = (
                "SUCCESS" if status in [None, "PENDING_VALIDATION"] else status
            )
            option["ValidationEmails"] = option.get("ValidationEmails") or [
                "admin@%s" % self.common_name
            ]
            test_record = {
                "Name": "test.%s" % domain_name,
                "Type": "CNAME",
                "Value": "test123",
            }
            option["ResourceRecord"] = option.get("ResourceRecord") or test_record

        for key, value in addenda.items():
            if not cert.get(key):
                cert[key] = value
        cert["Serial"] = str(cert.get("Serial") or "")
        return result

    describe_orig = acm_models.CertBundle.describe
    acm_models.CertBundle.describe = describe

    def wrap_request_certificate(backend):
        def request_certificate(self, domain_name, domain_validation_options, *args, **kwargs):
            cert_arn = request_certificate_orig(
                domain_name, domain_validation_options, *args, **kwargs
            )
            cert = self._certificates[cert_arn]
            if not hasattr(cert, "domain_validation_options"):
                cert.domain_validation_options = domain_validation_options
            return cert_arn

        request_certificate_orig = backend.request_certificate
        backend.request_certificate = types.MethodType(request_certificate, backend)

    for _, backend in acm_models.acm_backends.items():
        wrap_request_certificate(backend)


def start_acm(port=None, asynchronous=False, update_listener=None):
    port = port or config.service_port("acm")
    apply_patches()
    return start_moto_server(
        "acm",
        port,
        name="ACM",
        update_listener=update_listener,
        asynchronous=asynchronous,
    )



import sys
from datetime import datetime
from typing import Dict, List, Optional

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from localstack.aws.api import RequestContext, ServiceException, ServiceRequest, handler

AWSKMSKeyARN = str
AmazonopensearchserviceBufferingIntervalInSeconds = int
AmazonopensearchserviceBufferingSizeInMBs = int
AmazonopensearchserviceClusterEndpoint = str
AmazonopensearchserviceDomainARN = str
AmazonopensearchserviceIndexName = str
AmazonopensearchserviceRetryDurationInSeconds = int
AmazonopensearchserviceTypeName = str
BlockSizeBytes = int
BooleanObject = bool
BucketARN = str
ClusterJDBCURL = str
CopyOptions = str
DataTableColumns = str
DataTableName = str
DeliveryStreamARN = str
DeliveryStreamName = str
DeliveryStreamVersionId = str
DescribeDeliveryStreamInputLimit = int
DestinationId = str
ElasticsearchBufferingIntervalInSeconds = int
ElasticsearchBufferingSizeInMBs = int
ElasticsearchClusterEndpoint = str
ElasticsearchDomainARN = str
ElasticsearchIndexName = str
ElasticsearchRetryDurationInSeconds = int
ElasticsearchTypeName = str
ErrorCode = str
ErrorMessage = str
ErrorOutputPrefix = str
HECAcknowledgmentTimeoutInSeconds = int
HECEndpoint = str
HECToken = str
HttpEndpointAccessKey = str
HttpEndpointAttributeName = str
HttpEndpointAttributeValue = str
HttpEndpointBufferingIntervalInSeconds = int
HttpEndpointBufferingSizeInMBs = int
HttpEndpointName = str
HttpEndpointRetryDurationInSeconds = int
HttpEndpointUrl = str
IntervalInSeconds = int
KinesisStreamARN = str
ListDeliveryStreamsInputLimit = int
ListTagsForDeliveryStreamInputLimit = int
LogGroupName = str
LogStreamName = str
NonEmptyString = str
NonEmptyStringWithoutWhitespace = str
NonNegativeIntegerObject = int
OrcRowIndexStride = int
OrcStripeSizeBytes = int
ParquetPageSizeBytes = int
Password = str
Prefix = str
ProcessorParameterValue = str
Proportion = float
PutResponseRecordId = str
RedshiftRetryDurationInSeconds = int
RetryDurationInSeconds = int
RoleARN = str
SizeInMBs = int
SplunkRetryDurationInSeconds = int
TagKey = str
TagValue = str
Username = str


class AmazonopensearchserviceIndexRotationPeriod(str):
    NoRotation = "NoRotation"
    OneHour = "OneHour"
    OneDay = "OneDay"
    OneWeek = "OneWeek"
    OneMonth = "OneMonth"


class AmazonopensearchserviceS3BackupMode(str):
    FailedDocumentsOnly = "FailedDocumentsOnly"
    AllDocuments = "AllDocuments"


class CompressionFormat(str):
    UNCOMPRESSED = "UNCOMPRESSED"
    GZIP = "GZIP"
    ZIP = "ZIP"
    Snappy = "Snappy"
    HADOOP_SNAPPY = "HADOOP_SNAPPY"


class ContentEncoding(str):
    NONE = "NONE"
    GZIP = "GZIP"


class DeliveryStreamEncryptionStatus(str):
    ENABLED = "ENABLED"
    ENABLING = "ENABLING"
    ENABLING_FAILED = "ENABLING_FAILED"
    DISABLED = "DISABLED"
    DISABLING = "DISABLING"
    DISABLING_FAILED = "DISABLING_FAILED"


class DeliveryStreamFailureType(str):
    RETIRE_KMS_GRANT_FAILED = "RETIRE_KMS_GRANT_FAILED"
    CREATE_KMS_GRANT_FAILED = "CREATE_KMS_GRANT_FAILED"
    KMS_ACCESS_DENIED = "KMS_ACCESS_DENIED"
    DISABLED_KMS_KEY = "DISABLED_KMS_KEY"
    INVALID_KMS_KEY = "INVALID_KMS_KEY"
    KMS_KEY_NOT_FOUND = "KMS_KEY_NOT_FOUND"
    KMS_OPT_IN_REQUIRED = "KMS_OPT_IN_REQUIRED"
    CREATE_ENI_FAILED = "CREATE_ENI_FAILED"
    DELETE_ENI_FAILED = "DELETE_ENI_FAILED"
    SUBNET_NOT_FOUND = "SUBNET_NOT_FOUND"
    SECURITY_GROUP_NOT_FOUND = "SECURITY_GROUP_NOT_FOUND"
    ENI_ACCESS_DENIED = "ENI_ACCESS_DENIED"
    SUBNET_ACCESS_DENIED = "SUBNET_ACCESS_DENIED"
    SECURITY_GROUP_ACCESS_DENIED = "SECURITY_GROUP_ACCESS_DENIED"
    UNKNOWN_ERROR = "UNKNOWN_ERROR"


class DeliveryStreamStatus(str):
    CREATING = "CREATING"
    CREATING_FAILED = "CREATING_FAILED"
    DELETING = "DELETING"
    DELETING_FAILED = "DELETING_FAILED"
    ACTIVE = "ACTIVE"


class DeliveryStreamType(str):
    DirectPut = "DirectPut"
    KinesisStreamAsSource = "KinesisStreamAsSource"


class ElasticsearchIndexRotationPeriod(str):
    NoRotation = "NoRotation"
    OneHour = "OneHour"
    OneDay = "OneDay"
    OneWeek = "OneWeek"
    OneMonth = "OneMonth"


class ElasticsearchS3BackupMode(str):
    FailedDocumentsOnly = "FailedDocumentsOnly"
    AllDocuments = "AllDocuments"


class HECEndpointType(str):
    Raw = "Raw"
    Event = "Event"


class HttpEndpointS3BackupMode(str):
    FailedDataOnly = "FailedDataOnly"
    AllData = "AllData"


class KeyType(str):
    AWS_OWNED_CMK = "AWS_OWNED_CMK"
    CUSTOMER_MANAGED_CMK = "CUSTOMER_MANAGED_CMK"


class NoEncryptionConfig(str):
    NoEncryption = "NoEncryption"


class OrcCompression(str):
    NONE = "NONE"
    ZLIB = "ZLIB"
    SNAPPY = "SNAPPY"


class OrcFormatVersion(str):
    V0_11 = "V0_11"
    V0_12 = "V0_12"


class ParquetCompression(str):
    UNCOMPRESSED = "UNCOMPRESSED"
    GZIP = "GZIP"
    SNAPPY = "SNAPPY"


class ParquetWriterVersion(str):
    V1 = "V1"
    V2 = "V2"


class ProcessorParameterName(str):
    LambdaArn = "LambdaArn"
    NumberOfRetries = "NumberOfRetries"
    MetadataExtractionQuery = "MetadataExtractionQuery"
    JsonParsingEngine = "JsonParsingEngine"
    RoleArn = "RoleArn"
    BufferSizeInMBs = "BufferSizeInMBs"
    BufferIntervalInSeconds = "BufferIntervalInSeconds"
    SubRecordType = "SubRecordType"
    Delimiter = "Delimiter"


class ProcessorType(str):
    RecordDeAggregation = "RecordDeAggregation"
    Lambda = "Lambda"
    MetadataExtraction = "MetadataExtraction"
    AppendDelimiterToRecord = "AppendDelimiterToRecord"


class RedshiftS3BackupMode(str):
    Disabled = "Disabled"
    Enabled = "Enabled"


class S3BackupMode(str):
    Disabled = "Disabled"
    Enabled = "Enabled"


class SplunkS3BackupMode(str):
    FailedEventsOnly = "FailedEventsOnly"
    AllEvents = "AllEvents"


class ConcurrentModificationException(ServiceException):
    """Another modification has already happened. Fetch ``VersionId`` again and
    use it to update the destination.
    """

    message: Optional[ErrorMessage]


class InvalidArgumentException(ServiceException):
    """The specified input parameter has a value that is not valid."""

    message: Optional[ErrorMessage]


class InvalidKMSResourceException(ServiceException):
    """Kinesis Data Firehose throws this exception when an attempt to put
    records or to start or stop delivery stream encryption fails. This
    happens when the KMS service throws one of the following exception
    types: ``AccessDeniedException``, ``InvalidStateException``,
    ``DisabledException``, or ``NotFoundException``.
    """

    code: Optional[ErrorCode]
    message: Optional[ErrorMessage]


class LimitExceededException(ServiceException):
    """You have already reached the limit for a requested resource."""

    message: Optional[ErrorMessage]


class ResourceInUseException(ServiceException):
    """The resource is already in use and not available for this operation."""

    message: Optional[ErrorMessage]


class ResourceNotFoundException(ServiceException):
    """The specified resource could not be found."""

    message: Optional[ErrorMessage]


class ServiceUnavailableException(ServiceException):
    """The service is unavailable. Back off and retry the operation. If you
    continue to see the exception, throughput limits for the delivery stream
    may have been exceeded. For more information about limits and how to
    request an increase, see `Amazon Kinesis Data Firehose
    Limits <https://docs.aws.amazon.com/firehose/latest/dev/limits.html>`__.
    """

    message: Optional[ErrorMessage]


class AmazonopensearchserviceBufferingHints(TypedDict, total=False):
    IntervalInSeconds: Optional[AmazonopensearchserviceBufferingIntervalInSeconds]
    SizeInMBs: Optional[AmazonopensearchserviceBufferingSizeInMBs]


SecurityGroupIdList = List[NonEmptyStringWithoutWhitespace]
SubnetIdList = List[NonEmptyStringWithoutWhitespace]


class VpcConfiguration(TypedDict, total=False):
    """The details of the VPC of the Amazon ES destination."""

    SubnetIds: SubnetIdList
    RoleARN: RoleARN
    SecurityGroupIds: SecurityGroupIdList


class CloudWatchLoggingOptions(TypedDict, total=False):
    """Describes the Amazon CloudWatch logging options for your delivery
    stream.
    """

    Enabled: Optional[BooleanObject]
    LogGroupName: Optional[LogGroupName]
    LogStreamName: Optional[LogStreamName]


class ProcessorParameter(TypedDict, total=False):
    """Describes the processor parameter."""

    ParameterName: ProcessorParameterName
    ParameterValue: ProcessorParameterValue


ProcessorParameterList = List[ProcessorParameter]


class Processor(TypedDict, total=False):
    """Describes a data processor."""

    Type: ProcessorType
    Parameters: Optional[ProcessorParameterList]


ProcessorList = List[Processor]


class ProcessingConfiguration(TypedDict, total=False):
    """Describes a data processing configuration."""

    Enabled: Optional[BooleanObject]
    Processors: Optional[ProcessorList]


class KMSEncryptionConfig(TypedDict, total=False):
    """Describes an encryption key for a destination in Amazon S3."""

    AWSKMSKeyARN: AWSKMSKeyARN


class EncryptionConfiguration(TypedDict, total=False):
    """Describes the encryption for a destination in Amazon S3."""

    NoEncryptionConfig: Optional[NoEncryptionConfig]
    KMSEncryptionConfig: Optional[KMSEncryptionConfig]


class BufferingHints(TypedDict, total=False):
    """Describes hints for the buffering to perform before delivering data to
    the destination. These options are treated as hints, and therefore
    Kinesis Data Firehose might choose to use different values when it is
    optimal. The ``SizeInMBs`` and ``IntervalInSeconds`` parameters are
    optional. However, if specify a value for one of them, you must also
    provide a value for the other.
    """

    SizeInMBs: Optional[SizeInMBs]
    IntervalInSeconds: Optional[IntervalInSeconds]


class S3DestinationConfiguration(TypedDict, total=False):
    """Describes the configuration of a destination in Amazon S3."""

    RoleARN: RoleARN
    BucketARN: BucketARN
    Prefix: Optional[Prefix]
    ErrorOutputPrefix: Optional[ErrorOutputPrefix]
    BufferingHints: Optional[BufferingHints]
    CompressionFormat: Optional[CompressionFormat]
    EncryptionConfiguration: Optional[EncryptionConfiguration]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]


class AmazonopensearchserviceRetryOptions(TypedDict, total=False):
    DurationInSeconds: Optional[AmazonopensearchserviceRetryDurationInSeconds]


class AmazonopensearchserviceDestinationConfiguration(TypedDict, total=False):
    RoleARN: RoleARN
    DomainARN: Optional[AmazonopensearchserviceDomainARN]
    ClusterEndpoint: Optional[AmazonopensearchserviceClusterEndpoint]
    IndexName: AmazonopensearchserviceIndexName
    TypeName: Optional[AmazonopensearchserviceTypeName]
    IndexRotationPeriod: Optional[AmazonopensearchserviceIndexRotationPeriod]
    BufferingHints: Optional[AmazonopensearchserviceBufferingHints]
    RetryOptions: Optional[AmazonopensearchserviceRetryOptions]
    S3BackupMode: Optional[AmazonopensearchserviceS3BackupMode]
    S3Configuration: S3DestinationConfiguration
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]
    VpcConfiguration: Optional[VpcConfiguration]


class VpcConfigurationDescription(TypedDict, total=False):
    """The details of the VPC of the Amazon ES destination."""

    SubnetIds: SubnetIdList
    RoleARN: RoleARN
    SecurityGroupIds: SecurityGroupIdList
    VpcId: NonEmptyStringWithoutWhitespace


class S3DestinationDescription(TypedDict, total=False):
    """Describes a destination in Amazon S3."""

    RoleARN: RoleARN
    BucketARN: BucketARN
    Prefix: Optional[Prefix]
    ErrorOutputPrefix: Optional[ErrorOutputPrefix]
    BufferingHints: BufferingHints
    CompressionFormat: CompressionFormat
    EncryptionConfiguration: EncryptionConfiguration
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]


class AmazonopensearchserviceDestinationDescription(TypedDict, total=False):
    RoleARN: Optional[RoleARN]
    DomainARN: Optional[AmazonopensearchserviceDomainARN]
    ClusterEndpoint: Optional[AmazonopensearchserviceClusterEndpoint]
    IndexName: Optional[AmazonopensearchserviceIndexName]
    TypeName: Optional[AmazonopensearchserviceTypeName]
    IndexRotationPeriod: Optional[AmazonopensearchserviceIndexRotationPeriod]
    BufferingHints: Optional[AmazonopensearchserviceBufferingHints]
    RetryOptions: Optional[AmazonopensearchserviceRetryOptions]
    S3BackupMode: Optional[AmazonopensearchserviceS3BackupMode]
    S3DestinationDescription: Optional[S3DestinationDescription]
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]
    VpcConfigurationDescription: Optional[VpcConfigurationDescription]


class S3DestinationUpdate(TypedDict, total=False):
    """Describes an update for a destination in Amazon S3."""

    RoleARN: Optional[RoleARN]
    BucketARN: Optional[BucketARN]
    Prefix: Optional[Prefix]
    ErrorOutputPrefix: Optional[ErrorOutputPrefix]
    BufferingHints: Optional[BufferingHints]
    CompressionFormat: Optional[CompressionFormat]
    EncryptionConfiguration: Optional[EncryptionConfiguration]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]


class AmazonopensearchserviceDestinationUpdate(TypedDict, total=False):
    RoleARN: Optional[RoleARN]
    DomainARN: Optional[AmazonopensearchserviceDomainARN]
    ClusterEndpoint: Optional[AmazonopensearchserviceClusterEndpoint]
    IndexName: Optional[AmazonopensearchserviceIndexName]
    TypeName: Optional[AmazonopensearchserviceTypeName]
    IndexRotationPeriod: Optional[AmazonopensearchserviceIndexRotationPeriod]
    BufferingHints: Optional[AmazonopensearchserviceBufferingHints]
    RetryOptions: Optional[AmazonopensearchserviceRetryOptions]
    S3Update: Optional[S3DestinationUpdate]
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]


ColumnToJsonKeyMappings = Dict[NonEmptyStringWithoutWhitespace, NonEmptyString]


class CopyCommand(TypedDict, total=False):
    """Describes a ``COPY`` command for Amazon Redshift."""

    DataTableName: DataTableName
    DataTableColumns: Optional[DataTableColumns]
    CopyOptions: Optional[CopyOptions]


class Tag(TypedDict, total=False):
    """Metadata that you can assign to a delivery stream, consisting of a
    key-value pair.
    """

    Key: TagKey
    Value: Optional[TagValue]


TagDeliveryStreamInputTagList = List[Tag]


class HttpEndpointRetryOptions(TypedDict, total=False):
    """Describes the retry behavior in case Kinesis Data Firehose is unable to
    deliver data to the specified HTTP endpoint destination, or if it
    doesn't receive a valid acknowledgment of receipt from the specified
    HTTP endpoint destination.
    """

    DurationInSeconds: Optional[HttpEndpointRetryDurationInSeconds]


class HttpEndpointCommonAttribute(TypedDict, total=False):
    """Describes the metadata that's delivered to the specified HTTP endpoint
    destination.
    """

    AttributeName: HttpEndpointAttributeName
    AttributeValue: HttpEndpointAttributeValue


HttpEndpointCommonAttributesList = List[HttpEndpointCommonAttribute]


class HttpEndpointRequestConfiguration(TypedDict, total=False):
    """The configuration of the HTTP endpoint request."""

    ContentEncoding: Optional[ContentEncoding]
    CommonAttributes: Optional[HttpEndpointCommonAttributesList]


class HttpEndpointBufferingHints(TypedDict, total=False):
    """Describes the buffering options that can be applied before data is
    delivered to the HTTP endpoint destination. Kinesis Data Firehose treats
    these options as hints, and it might choose to use more optimal values.
    The ``SizeInMBs`` and ``IntervalInSeconds`` parameters are optional.
    However, if specify a value for one of them, you must also provide a
    value for the other.
    """

    SizeInMBs: Optional[HttpEndpointBufferingSizeInMBs]
    IntervalInSeconds: Optional[HttpEndpointBufferingIntervalInSeconds]


class HttpEndpointConfiguration(TypedDict, total=False):
    """Describes the configuration of the HTTP endpoint to which Kinesis
    Firehose delivers data.
    """

    Url: HttpEndpointUrl
    Name: Optional[HttpEndpointName]
    AccessKey: Optional[HttpEndpointAccessKey]


class HttpEndpointDestinationConfiguration(TypedDict, total=False):
    """Describes the configuration of the HTTP endpoint destination."""

    EndpointConfiguration: HttpEndpointConfiguration
    BufferingHints: Optional[HttpEndpointBufferingHints]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]
    RequestConfiguration: Optional[HttpEndpointRequestConfiguration]
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    RoleARN: Optional[RoleARN]
    RetryOptions: Optional[HttpEndpointRetryOptions]
    S3BackupMode: Optional[HttpEndpointS3BackupMode]
    S3Configuration: S3DestinationConfiguration


class SplunkRetryOptions(TypedDict, total=False):
    """Configures retry behavior in case Kinesis Data Firehose is unable to
    deliver documents to Splunk, or if it doesn't receive an acknowledgment
    from Splunk.
    """

    DurationInSeconds: Optional[SplunkRetryDurationInSeconds]


class SplunkDestinationConfiguration(TypedDict, total=False):
    """Describes the configuration of a destination in Splunk."""

    HECEndpoint: HECEndpoint
    HECEndpointType: HECEndpointType
    HECToken: HECToken
    HECAcknowledgmentTimeoutInSeconds: Optional[HECAcknowledgmentTimeoutInSeconds]
    RetryOptions: Optional[SplunkRetryOptions]
    S3BackupMode: Optional[SplunkS3BackupMode]
    S3Configuration: S3DestinationConfiguration
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]


class ElasticsearchRetryOptions(TypedDict, total=False):
    """Configures retry behavior in case Kinesis Data Firehose is unable to
    deliver documents to Amazon ES.
    """

    DurationInSeconds: Optional[ElasticsearchRetryDurationInSeconds]


class ElasticsearchBufferingHints(TypedDict, total=False):
    """Describes the buffering to perform before delivering data to the Amazon
    ES destination.
    """

    IntervalInSeconds: Optional[ElasticsearchBufferingIntervalInSeconds]
    SizeInMBs: Optional[ElasticsearchBufferingSizeInMBs]


class ElasticsearchDestinationConfiguration(TypedDict, total=False):
    """Describes the configuration of a destination in Amazon ES."""

    RoleARN: RoleARN
    DomainARN: Optional[ElasticsearchDomainARN]
    ClusterEndpoint: Optional[ElasticsearchClusterEndpoint]
    IndexName: ElasticsearchIndexName
    TypeName: Optional[ElasticsearchTypeName]
    IndexRotationPeriod: Optional[ElasticsearchIndexRotationPeriod]
    BufferingHints: Optional[ElasticsearchBufferingHints]
    RetryOptions: Optional[ElasticsearchRetryOptions]
    S3BackupMode: Optional[ElasticsearchS3BackupMode]
    S3Configuration: S3DestinationConfiguration
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]
    VpcConfiguration: Optional[VpcConfiguration]


class RedshiftRetryOptions(TypedDict, total=False):
    """Configures retry behavior in case Kinesis Data Firehose is unable to
    deliver documents to Amazon Redshift.
    """

    DurationInSeconds: Optional[RedshiftRetryDurationInSeconds]


class RedshiftDestinationConfiguration(TypedDict, total=False):
    """Describes the configuration of a destination in Amazon Redshift."""

    RoleARN: RoleARN
    ClusterJDBCURL: ClusterJDBCURL
    CopyCommand: CopyCommand
    Username: Username
    Password: Password
    RetryOptions: Optional[RedshiftRetryOptions]
    S3Configuration: S3DestinationConfiguration
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    S3BackupMode: Optional[RedshiftS3BackupMode]
    S3BackupConfiguration: Optional[S3DestinationConfiguration]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]


class RetryOptions(TypedDict, total=False):
    """The retry behavior in case Kinesis Data Firehose is unable to deliver
    data to an Amazon S3 prefix.
    """

    DurationInSeconds: Optional[RetryDurationInSeconds]


class DynamicPartitioningConfiguration(TypedDict, total=False):
    """The configuration of the dynamic partitioning mechanism that creates
    smaller data sets from the streaming data by partitioning it based on
    partition keys. Currently, dynamic partitioning is only supported for
    Amazon S3 destinations. For more information, see
    https://docs.aws.amazon.com/firehose/latest/dev/dynamic-partitioning.html
    """

    RetryOptions: Optional[RetryOptions]
    Enabled: Optional[BooleanObject]


ListOfNonEmptyStringsWithoutWhitespace = List[NonEmptyStringWithoutWhitespace]


class OrcSerDe(TypedDict, total=False):
    """A serializer to use for converting data to the ORC format before storing
    it in Amazon S3. For more information, see `Apache
    ORC <https://orc.apache.org/docs/>`__.
    """

    StripeSizeBytes: Optional[OrcStripeSizeBytes]
    BlockSizeBytes: Optional[BlockSizeBytes]
    RowIndexStride: Optional[OrcRowIndexStride]
    EnablePadding: Optional[BooleanObject]
    PaddingTolerance: Optional[Proportion]
    Compression: Optional[OrcCompression]
    BloomFilterColumns: Optional[ListOfNonEmptyStringsWithoutWhitespace]
    BloomFilterFalsePositiveProbability: Optional[Proportion]
    DictionaryKeyThreshold: Optional[Proportion]
    FormatVersion: Optional[OrcFormatVersion]


class ParquetSerDe(TypedDict, total=False):
    """A serializer to use for converting data to the Parquet format before
    storing it in Amazon S3. For more information, see `Apache
    Parquet <https://parquet.apache.org/documentation/latest/>`__.
    """

    BlockSizeBytes: Optional[BlockSizeBytes]
    PageSizeBytes: Optional[ParquetPageSizeBytes]
    Compression: Optional[ParquetCompression]
    EnableDictionaryCompression: Optional[BooleanObject]
    MaxPaddingBytes: Optional[NonNegativeIntegerObject]
    WriterVersion: Optional[ParquetWriterVersion]


class Serializer(TypedDict, total=False):
    """The serializer that you want Kinesis Data Firehose to use to convert
    data to the target format before writing it to Amazon S3. Kinesis Data
    Firehose supports two types of serializers: the `ORC
    SerDe <https://hive.apache.org/javadocs/r1.2.2/api/org/apache/hadoop/hive/ql/io/orc/OrcSerde.html>`__
    and the `Parquet
    SerDe <https://hive.apache.org/javadocs/r1.2.2/api/org/apache/hadoop/hive/ql/io/parquet/serde/ParquetHiveSerDe.html>`__.
    """

    ParquetSerDe: Optional[ParquetSerDe]
    OrcSerDe: Optional[OrcSerDe]


class OutputFormatConfiguration(TypedDict, total=False):
    """Specifies the serializer that you want Kinesis Data Firehose to use to
    convert the format of your data before it writes it to Amazon S3. This
    parameter is required if ``Enabled`` is set to true.
    """

    Serializer: Optional[Serializer]


ListOfNonEmptyStrings = List[NonEmptyString]


class HiveJsonSerDe(TypedDict, total=False):
    """The native Hive / HCatalog JsonSerDe. Used by Kinesis Data Firehose for
    deserializing data, which means converting it from the JSON format in
    preparation for serializing it to the Parquet or ORC format. This is one
    of two deserializers you can choose, depending on which one offers the
    functionality you need. The other option is the OpenX SerDe.
    """

    TimestampFormats: Optional[ListOfNonEmptyStrings]


class OpenXJsonSerDe(TypedDict, total=False):
    """The OpenX SerDe. Used by Kinesis Data Firehose for deserializing data,
    which means converting it from the JSON format in preparation for
    serializing it to the Parquet or ORC format. This is one of two
    deserializers you can choose, depending on which one offers the
    functionality you need. The other option is the native Hive / HCatalog
    JsonSerDe.
    """

    ConvertDotsInJsonKeysToUnderscores: Optional[BooleanObject]
    CaseInsensitive: Optional[BooleanObject]
    ColumnToJsonKeyMappings: Optional[ColumnToJsonKeyMappings]


class Deserializer(TypedDict, total=False):
    """The deserializer you want Kinesis Data Firehose to use for converting
    the input data from JSON. Kinesis Data Firehose then serializes the data
    to its final format using the Serializer. Kinesis Data Firehose supports
    two types of deserializers: the `Apache Hive JSON
    SerDe <https://cwiki.apache.org/confluence/display/Hive/LanguageManual+DDL#LanguageManualDDL-JSON>`__
    and the `OpenX JSON
    SerDe <https://github.com/rcongiu/Hive-JSON-Serde>`__.
    """

    OpenXJsonSerDe: Optional[OpenXJsonSerDe]
    HiveJsonSerDe: Optional[HiveJsonSerDe]


class InputFormatConfiguration(TypedDict, total=False):
    """Specifies the deserializer you want to use to convert the format of the
    input data. This parameter is required if ``Enabled`` is set to true.
    """

    Deserializer: Optional[Deserializer]


class SchemaConfiguration(TypedDict, total=False):
    """Specifies the schema to which you want Kinesis Data Firehose to
    configure your data before it writes it to Amazon S3. This parameter is
    required if ``Enabled`` is set to true.
    """

    RoleARN: Optional[NonEmptyStringWithoutWhitespace]
    CatalogId: Optional[NonEmptyStringWithoutWhitespace]
    DatabaseName: Optional[NonEmptyStringWithoutWhitespace]
    TableName: Optional[NonEmptyStringWithoutWhitespace]
    Region: Optional[NonEmptyStringWithoutWhitespace]
    VersionId: Optional[NonEmptyStringWithoutWhitespace]


class DataFormatConversionConfiguration(TypedDict, total=False):
    """Specifies that you want Kinesis Data Firehose to convert data from the
    JSON format to the Parquet or ORC format before writing it to Amazon S3.
    Kinesis Data Firehose uses the serializer and deserializer that you
    specify, in addition to the column information from the AWS Glue table,
    to deserialize your input data from JSON and then serialize it to the
    Parquet or ORC format. For more information, see `Kinesis Data Firehose
    Record Format
    Conversion <https://docs.aws.amazon.com/firehose/latest/dev/record-format-conversion.html>`__.
    """

    SchemaConfiguration: Optional[SchemaConfiguration]
    InputFormatConfiguration: Optional[InputFormatConfiguration]
    OutputFormatConfiguration: Optional[OutputFormatConfiguration]
    Enabled: Optional[BooleanObject]


class ExtendedS3DestinationConfiguration(TypedDict, total=False):
    """Describes the configuration of a destination in Amazon S3."""

    RoleARN: RoleARN
    BucketARN: BucketARN
    Prefix: Optional[Prefix]
    ErrorOutputPrefix: Optional[ErrorOutputPrefix]
    BufferingHints: Optional[BufferingHints]
    CompressionFormat: Optional[CompressionFormat]
    EncryptionConfiguration: Optional[EncryptionConfiguration]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    S3BackupMode: Optional[S3BackupMode]
    S3BackupConfiguration: Optional[S3DestinationConfiguration]
    DataFormatConversionConfiguration: Optional[DataFormatConversionConfiguration]
    DynamicPartitioningConfiguration: Optional[DynamicPartitioningConfiguration]


class DeliveryStreamEncryptionConfigurationInput(TypedDict, total=False):
    """Specifies the type and Amazon Resource Name (ARN) of the CMK to use for
    Server-Side Encryption (SSE).
    """

    KeyARN: Optional[AWSKMSKeyARN]
    KeyType: KeyType


class KinesisStreamSourceConfiguration(TypedDict, total=False):
    """The stream and role Amazon Resource Names (ARNs) for a Kinesis data
    stream used as the source for a delivery stream.
    """

    KinesisStreamARN: KinesisStreamARN
    RoleARN: RoleARN


class CreateDeliveryStreamInput(ServiceRequest):
    DeliveryStreamName: DeliveryStreamName
    DeliveryStreamType: Optional[DeliveryStreamType]
    KinesisStreamSourceConfiguration: Optional[KinesisStreamSourceConfiguration]
    DeliveryStreamEncryptionConfigurationInput: Optional[DeliveryStreamEncryptionConfigurationInput]
    S3DestinationConfiguration: Optional[S3DestinationConfiguration]
    ExtendedS3DestinationConfiguration: Optional[ExtendedS3DestinationConfiguration]
    RedshiftDestinationConfiguration: Optional[RedshiftDestinationConfiguration]
    ElasticsearchDestinationConfiguration: Optional[ElasticsearchDestinationConfiguration]
    AmazonopensearchserviceDestinationConfiguration: Optional[
        AmazonopensearchserviceDestinationConfiguration
    ]
    SplunkDestinationConfiguration: Optional[SplunkDestinationConfiguration]
    HttpEndpointDestinationConfiguration: Optional[HttpEndpointDestinationConfiguration]
    Tags: Optional[TagDeliveryStreamInputTagList]


class CreateDeliveryStreamOutput(TypedDict, total=False):
    DeliveryStreamARN: Optional[DeliveryStreamARN]


Data = bytes


class DeleteDeliveryStreamInput(ServiceRequest):
    DeliveryStreamName: DeliveryStreamName
    AllowForceDelete: Optional[BooleanObject]


class DeleteDeliveryStreamOutput(TypedDict, total=False):
    pass


DeliveryStartTimestamp = datetime


class HttpEndpointDescription(TypedDict, total=False):
    """Describes the HTTP endpoint selected as the destination."""

    Url: Optional[HttpEndpointUrl]
    Name: Optional[HttpEndpointName]


class HttpEndpointDestinationDescription(TypedDict, total=False):
    """Describes the HTTP endpoint destination."""

    EndpointConfiguration: Optional[HttpEndpointDescription]
    BufferingHints: Optional[HttpEndpointBufferingHints]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]
    RequestConfiguration: Optional[HttpEndpointRequestConfiguration]
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    RoleARN: Optional[RoleARN]
    RetryOptions: Optional[HttpEndpointRetryOptions]
    S3BackupMode: Optional[HttpEndpointS3BackupMode]
    S3DestinationDescription: Optional[S3DestinationDescription]


class SplunkDestinationDescription(TypedDict, total=False):
    """Describes a destination in Splunk."""

    HECEndpoint: Optional[HECEndpoint]
    HECEndpointType: Optional[HECEndpointType]
    HECToken: Optional[HECToken]
    HECAcknowledgmentTimeoutInSeconds: Optional[HECAcknowledgmentTimeoutInSeconds]
    RetryOptions: Optional[SplunkRetryOptions]
    S3BackupMode: Optional[SplunkS3BackupMode]
    S3DestinationDescription: Optional[S3DestinationDescription]
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]


class ElasticsearchDestinationDescription(TypedDict, total=False):
    """The destination description in Amazon ES."""

    RoleARN: Optional[RoleARN]
    DomainARN: Optional[ElasticsearchDomainARN]
    ClusterEndpoint: Optional[ElasticsearchClusterEndpoint]
    IndexName: Optional[ElasticsearchIndexName]
    TypeName: Optional[ElasticsearchTypeName]
    IndexRotationPeriod: Optional[ElasticsearchIndexRotationPeriod]
    BufferingHints: Optional[ElasticsearchBufferingHints]
    RetryOptions: Optional[ElasticsearchRetryOptions]
    S3BackupMode: Optional[ElasticsearchS3BackupMode]
    S3DestinationDescription: Optional[S3DestinationDescription]
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]
    VpcConfigurationDescription: Optional[VpcConfigurationDescription]


class RedshiftDestinationDescription(TypedDict, total=False):
    """Describes a destination in Amazon Redshift."""

    RoleARN: RoleARN
    ClusterJDBCURL: ClusterJDBCURL
    CopyCommand: CopyCommand
    Username: Username
    RetryOptions: Optional[RedshiftRetryOptions]
    S3DestinationDescription: S3DestinationDescription
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    S3BackupMode: Optional[RedshiftS3BackupMode]
    S3BackupDescription: Optional[S3DestinationDescription]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]


class ExtendedS3DestinationDescription(TypedDict, total=False):
    """Describes a destination in Amazon S3."""

    RoleARN: RoleARN
    BucketARN: BucketARN
    Prefix: Optional[Prefix]
    ErrorOutputPrefix: Optional[ErrorOutputPrefix]
    BufferingHints: BufferingHints
    CompressionFormat: CompressionFormat
    EncryptionConfiguration: EncryptionConfiguration
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    S3BackupMode: Optional[S3BackupMode]
    S3BackupDescription: Optional[S3DestinationDescription]
    DataFormatConversionConfiguration: Optional[DataFormatConversionConfiguration]
    DynamicPartitioningConfiguration: Optional[DynamicPartitioningConfiguration]


class DestinationDescription(TypedDict, total=False):
    """Describes the destination for a delivery stream."""

    DestinationId: DestinationId
    S3DestinationDescription: Optional[S3DestinationDescription]
    ExtendedS3DestinationDescription: Optional[ExtendedS3DestinationDescription]
    RedshiftDestinationDescription: Optional[RedshiftDestinationDescription]
    ElasticsearchDestinationDescription: Optional[ElasticsearchDestinationDescription]
    AmazonopensearchserviceDestinationDescription: Optional[
        AmazonopensearchserviceDestinationDescription
    ]
    SplunkDestinationDescription: Optional[SplunkDestinationDescription]
    HttpEndpointDestinationDescription: Optional[HttpEndpointDestinationDescription]


DestinationDescriptionList = List[DestinationDescription]


class KinesisStreamSourceDescription(TypedDict, total=False):
    """Details about a Kinesis data stream used as the source for a Kinesis
    Data Firehose delivery stream.
    """

    KinesisStreamARN: Optional[KinesisStreamARN]
    RoleARN: Optional[RoleARN]
    DeliveryStartTimestamp: Optional[DeliveryStartTimestamp]


class SourceDescription(TypedDict, total=False):
    """Details about a Kinesis data stream used as the source for a Kinesis
    Data Firehose delivery stream.
    """

    KinesisStreamSourceDescription: Optional[KinesisStreamSourceDescription]


Timestamp = datetime


class FailureDescription(TypedDict, total=False):
    """Provides details in case one of the following operations fails due to an
    error related to KMS: CreateDeliveryStream, DeleteDeliveryStream,
    StartDeliveryStreamEncryption, StopDeliveryStreamEncryption.
    """

    Type: DeliveryStreamFailureType
    Details: NonEmptyString


class DeliveryStreamEncryptionConfiguration(TypedDict, total=False):
    """Contains information about the server-side encryption (SSE) status for
    the delivery stream, the type customer master key (CMK) in use, if any,
    and the ARN of the CMK. You can get
    ``DeliveryStreamEncryptionConfiguration`` by invoking the
    DescribeDeliveryStream operation.
    """

    KeyARN: Optional[AWSKMSKeyARN]
    KeyType: Optional[KeyType]
    Status: Optional[DeliveryStreamEncryptionStatus]
    FailureDescription: Optional[FailureDescription]


class DeliveryStreamDescription(TypedDict, total=False):
    """Contains information about a delivery stream."""

    DeliveryStreamName: DeliveryStreamName
    DeliveryStreamARN: DeliveryStreamARN
    DeliveryStreamStatus: DeliveryStreamStatus
    FailureDescription: Optional[FailureDescription]
    DeliveryStreamEncryptionConfiguration: Optional[DeliveryStreamEncryptionConfiguration]
    DeliveryStreamType: DeliveryStreamType
    VersionId: DeliveryStreamVersionId
    CreateTimestamp: Optional[Timestamp]
    LastUpdateTimestamp: Optional[Timestamp]
    Source: Optional[SourceDescription]
    Destinations: DestinationDescriptionList
    HasMoreDestinations: BooleanObject


DeliveryStreamNameList = List[DeliveryStreamName]


class DescribeDeliveryStreamInput(ServiceRequest):
    DeliveryStreamName: DeliveryStreamName
    Limit: Optional[DescribeDeliveryStreamInputLimit]
    ExclusiveStartDestinationId: Optional[DestinationId]


class DescribeDeliveryStreamOutput(TypedDict, total=False):
    DeliveryStreamDescription: DeliveryStreamDescription


class ElasticsearchDestinationUpdate(TypedDict, total=False):
    """Describes an update for a destination in Amazon ES."""

    RoleARN: Optional[RoleARN]
    DomainARN: Optional[ElasticsearchDomainARN]
    ClusterEndpoint: Optional[ElasticsearchClusterEndpoint]
    IndexName: Optional[ElasticsearchIndexName]
    TypeName: Optional[ElasticsearchTypeName]
    IndexRotationPeriod: Optional[ElasticsearchIndexRotationPeriod]
    BufferingHints: Optional[ElasticsearchBufferingHints]
    RetryOptions: Optional[ElasticsearchRetryOptions]
    S3Update: Optional[S3DestinationUpdate]
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]


class ExtendedS3DestinationUpdate(TypedDict, total=False):
    """Describes an update for a destination in Amazon S3."""

    RoleARN: Optional[RoleARN]
    BucketARN: Optional[BucketARN]
    Prefix: Optional[Prefix]
    ErrorOutputPrefix: Optional[ErrorOutputPrefix]
    BufferingHints: Optional[BufferingHints]
    CompressionFormat: Optional[CompressionFormat]
    EncryptionConfiguration: Optional[EncryptionConfiguration]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    S3BackupMode: Optional[S3BackupMode]
    S3BackupUpdate: Optional[S3DestinationUpdate]
    DataFormatConversionConfiguration: Optional[DataFormatConversionConfiguration]
    DynamicPartitioningConfiguration: Optional[DynamicPartitioningConfiguration]


class HttpEndpointDestinationUpdate(TypedDict, total=False):
    """Updates the specified HTTP endpoint destination."""

    EndpointConfiguration: Optional[HttpEndpointConfiguration]
    BufferingHints: Optional[HttpEndpointBufferingHints]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]
    RequestConfiguration: Optional[HttpEndpointRequestConfiguration]
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    RoleARN: Optional[RoleARN]
    RetryOptions: Optional[HttpEndpointRetryOptions]
    S3BackupMode: Optional[HttpEndpointS3BackupMode]
    S3Update: Optional[S3DestinationUpdate]


class ListDeliveryStreamsInput(ServiceRequest):
    Limit: Optional[ListDeliveryStreamsInputLimit]
    DeliveryStreamType: Optional[DeliveryStreamType]
    ExclusiveStartDeliveryStreamName: Optional[DeliveryStreamName]


class ListDeliveryStreamsOutput(TypedDict, total=False):
    DeliveryStreamNames: DeliveryStreamNameList
    HasMoreDeliveryStreams: BooleanObject


class ListTagsForDeliveryStreamInput(ServiceRequest):
    DeliveryStreamName: DeliveryStreamName
    ExclusiveStartTagKey: Optional[TagKey]
    Limit: Optional[ListTagsForDeliveryStreamInputLimit]


ListTagsForDeliveryStreamOutputTagList = List[Tag]


class ListTagsForDeliveryStreamOutput(TypedDict, total=False):
    Tags: ListTagsForDeliveryStreamOutputTagList
    HasMoreTags: BooleanObject


class Record(TypedDict, total=False):
    """The unit of data in a delivery stream."""

    Data: Data


PutRecordBatchRequestEntryList = List[Record]


class PutRecordBatchInput(ServiceRequest):
    DeliveryStreamName: DeliveryStreamName
    Records: PutRecordBatchRequestEntryList


class PutRecordBatchResponseEntry(TypedDict, total=False):
    """Contains the result for an individual record from a PutRecordBatch
    request. If the record is successfully added to your delivery stream, it
    receives a record ID. If the record fails to be added to your delivery
    stream, the result includes an error code and an error message.
    """

    RecordId: Optional[PutResponseRecordId]
    ErrorCode: Optional[ErrorCode]
    ErrorMessage: Optional[ErrorMessage]


PutRecordBatchResponseEntryList = List[PutRecordBatchResponseEntry]


class PutRecordBatchOutput(TypedDict, total=False):
    FailedPutCount: NonNegativeIntegerObject
    Encrypted: Optional[BooleanObject]
    RequestResponses: PutRecordBatchResponseEntryList


class PutRecordInput(ServiceRequest):
    DeliveryStreamName: DeliveryStreamName
    Record: Record


class PutRecordOutput(TypedDict, total=False):
    RecordId: PutResponseRecordId
    Encrypted: Optional[BooleanObject]


class RedshiftDestinationUpdate(TypedDict, total=False):
    """Describes an update for a destination in Amazon Redshift."""

    RoleARN: Optional[RoleARN]
    ClusterJDBCURL: Optional[ClusterJDBCURL]
    CopyCommand: Optional[CopyCommand]
    Username: Optional[Username]
    Password: Optional[Password]
    RetryOptions: Optional[RedshiftRetryOptions]
    S3Update: Optional[S3DestinationUpdate]
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    S3BackupMode: Optional[RedshiftS3BackupMode]
    S3BackupUpdate: Optional[S3DestinationUpdate]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]


class SplunkDestinationUpdate(TypedDict, total=False):
    """Describes an update for a destination in Splunk."""

    HECEndpoint: Optional[HECEndpoint]
    HECEndpointType: Optional[HECEndpointType]
    HECToken: Optional[HECToken]
    HECAcknowledgmentTimeoutInSeconds: Optional[HECAcknowledgmentTimeoutInSeconds]
    RetryOptions: Optional[SplunkRetryOptions]
    S3BackupMode: Optional[SplunkS3BackupMode]
    S3Update: Optional[S3DestinationUpdate]
    ProcessingConfiguration: Optional[ProcessingConfiguration]
    CloudWatchLoggingOptions: Optional[CloudWatchLoggingOptions]


class StartDeliveryStreamEncryptionInput(ServiceRequest):
    DeliveryStreamName: DeliveryStreamName
    DeliveryStreamEncryptionConfigurationInput: Optional[DeliveryStreamEncryptionConfigurationInput]


class StartDeliveryStreamEncryptionOutput(TypedDict, total=False):
    pass


class StopDeliveryStreamEncryptionInput(ServiceRequest):
    DeliveryStreamName: DeliveryStreamName


class StopDeliveryStreamEncryptionOutput(TypedDict, total=False):
    pass


class TagDeliveryStreamInput(ServiceRequest):
    DeliveryStreamName: DeliveryStreamName
    Tags: TagDeliveryStreamInputTagList


class TagDeliveryStreamOutput(TypedDict, total=False):
    pass


TagKeyList = List[TagKey]


class UntagDeliveryStreamInput(ServiceRequest):
    DeliveryStreamName: DeliveryStreamName
    TagKeys: TagKeyList


class UntagDeliveryStreamOutput(TypedDict, total=False):
    pass


class UpdateDestinationInput(ServiceRequest):
    DeliveryStreamName: DeliveryStreamName
    CurrentDeliveryStreamVersionId: DeliveryStreamVersionId
    DestinationId: DestinationId
    S3DestinationUpdate: Optional[S3DestinationUpdate]
    ExtendedS3DestinationUpdate: Optional[ExtendedS3DestinationUpdate]
    RedshiftDestinationUpdate: Optional[RedshiftDestinationUpdate]
    ElasticsearchDestinationUpdate: Optional[ElasticsearchDestinationUpdate]
    AmazonopensearchserviceDestinationUpdate: Optional[AmazonopensearchserviceDestinationUpdate]
    SplunkDestinationUpdate: Optional[SplunkDestinationUpdate]
    HttpEndpointDestinationUpdate: Optional[HttpEndpointDestinationUpdate]


class UpdateDestinationOutput(TypedDict, total=False):
    pass


class FirehoseApi:

    service = "firehose"
    version = "2015-08-04"

    @handler("CreateDeliveryStream")
    def create_delivery_stream(
        self,
        context: RequestContext,
        delivery_stream_name: DeliveryStreamName,
        delivery_stream_type: DeliveryStreamType = None,
        kinesis_stream_source_configuration: KinesisStreamSourceConfiguration = None,
        delivery_stream_encryption_configuration_input: DeliveryStreamEncryptionConfigurationInput = None,
        s3_destination_configuration: S3DestinationConfiguration = None,
        extended_s3_destination_configuration: ExtendedS3DestinationConfiguration = None,
        redshift_destination_configuration: RedshiftDestinationConfiguration = None,
        elasticsearch_destination_configuration: ElasticsearchDestinationConfiguration = None,
        amazonopensearchservice_destination_configuration: AmazonopensearchserviceDestinationConfiguration = None,
        splunk_destination_configuration: SplunkDestinationConfiguration = None,
        http_endpoint_destination_configuration: HttpEndpointDestinationConfiguration = None,
        tags: TagDeliveryStreamInputTagList = None,
    ) -> CreateDeliveryStreamOutput:
        """Creates a Kinesis Data Firehose delivery stream.

        By default, you can create up to 50 delivery streams per AWS Region.

        This is an asynchronous operation that immediately returns. The initial
        status of the delivery stream is ``CREATING``. After the delivery stream
        is created, its status is ``ACTIVE`` and it now accepts data. If the
        delivery stream creation fails, the status transitions to
        ``CREATING_FAILED``. Attempts to send data to a delivery stream that is
        not in the ``ACTIVE`` state cause an exception. To check the state of a
        delivery stream, use DescribeDeliveryStream.

        If the status of a delivery stream is ``CREATING_FAILED``, this status
        doesn't change, and you can't invoke ``CreateDeliveryStream`` again on
        it. However, you can invoke the DeleteDeliveryStream operation to delete
        it.

        A Kinesis Data Firehose delivery stream can be configured to receive
        records directly from providers using PutRecord or PutRecordBatch, or it
        can be configured to use an existing Kinesis stream as its source. To
        specify a Kinesis data stream as input, set the ``DeliveryStreamType``
        parameter to ``KinesisStreamAsSource``, and provide the Kinesis stream
        Amazon Resource Name (ARN) and role ARN in the
        ``KinesisStreamSourceConfiguration`` parameter.

        To create a delivery stream with server-side encryption (SSE) enabled,
        include DeliveryStreamEncryptionConfigurationInput in your request. This
        is optional. You can also invoke StartDeliveryStreamEncryption to turn
        on SSE for an existing delivery stream that doesn't have SSE enabled.

        A delivery stream is configured with a single destination: Amazon S3,
        Amazon ES, Amazon Redshift, or Splunk. You must specify only one of the
        following destination configuration parameters:
        ``ExtendedS3DestinationConfiguration``, ``S3DestinationConfiguration``,
        ``ElasticsearchDestinationConfiguration``,
        ``RedshiftDestinationConfiguration``, or
        ``SplunkDestinationConfiguration``.

        When you specify ``S3DestinationConfiguration``, you can also provide
        the following optional values: BufferingHints,
        ``EncryptionConfiguration``, and ``CompressionFormat``. By default, if
        no ``BufferingHints`` value is provided, Kinesis Data Firehose buffers
        data up to 5 MB or for 5 minutes, whichever condition is satisfied
        first. ``BufferingHints`` is a hint, so there are some cases where the
        service cannot adhere to these conditions strictly. For example, record
        boundaries might be such that the size is a little over or under the
        configured buffering size. By default, no encryption is performed. We
        strongly recommend that you enable encryption to ensure secure data
        storage in Amazon S3.

        A few notes about Amazon Redshift as a destination:

        -  An Amazon Redshift destination requires an S3 bucket as intermediate
           location. Kinesis Data Firehose first delivers data to Amazon S3 and
           then uses ``COPY`` syntax to load data into an Amazon Redshift table.
           This is specified in the
           ``RedshiftDestinationConfiguration.S3Configuration`` parameter.

        -  The compression formats ``SNAPPY`` or ``ZIP`` cannot be specified in
           ``RedshiftDestinationConfiguration.S3Configuration`` because the
           Amazon Redshift ``COPY`` operation that reads from the S3 bucket
           doesn't support these compression formats.

        -  We strongly recommend that you use the user name and password you
           provide exclusively with Kinesis Data Firehose, and that the
           permissions for the account are restricted for Amazon Redshift
           ``INSERT`` permissions.

        Kinesis Data Firehose assumes the IAM role that is configured as part of
        the destination. The role should allow the Kinesis Data Firehose
        principal to assume the role, and the role should have permissions that
        allow the service to deliver the data. For more information, see `Grant
        Kinesis Data Firehose Access to an Amazon S3
        Destination <https://docs.aws.amazon.com/firehose/latest/dev/controlling-access.html#using-iam-s3>`__
        in the *Amazon Kinesis Data Firehose Developer Guide*.

        :param delivery_stream_name: The name of the delivery stream.
        :param delivery_stream_type: The delivery stream type.
        :param kinesis_stream_source_configuration: When a Kinesis data stream is used as the source for the delivery
        stream, a KinesisStreamSourceConfiguration containing the Kinesis data
        stream Amazon Resource Name (ARN) and the role ARN for the source
        stream.
        :param delivery_stream_encryption_configuration_input: Used to specify the type and Amazon Resource Name (ARN) of the KMS key
        needed for Server-Side Encryption (SSE).
        :param s3_destination_configuration: [Deprecated] The destination in Amazon S3.
        :param extended_s3_destination_configuration: The destination in Amazon S3.
        :param redshift_destination_configuration: The destination in Amazon Redshift.
        :param elasticsearch_destination_configuration: The destination in Amazon ES.
        :param amazonopensearchservice_destination_configuration: .
        :param splunk_destination_configuration: The destination in Splunk.
        :param http_endpoint_destination_configuration: Enables configuring Kinesis Firehose to deliver data to any HTTP
        endpoint destination.
        :param tags: A set of tags to assign to the delivery stream.
        :returns: CreateDeliveryStreamOutput
        :raises InvalidArgumentException:
        :raises LimitExceededException:
        :raises ResourceInUseException:
        :raises InvalidKMSResourceException:
        """
        raise NotImplementedError

    @handler("DeleteDeliveryStream")
    def delete_delivery_stream(
        self,
        context: RequestContext,
        delivery_stream_name: DeliveryStreamName,
        allow_force_delete: BooleanObject = None,
    ) -> DeleteDeliveryStreamOutput:
        """Deletes a delivery stream and its data.

        To check the state of a delivery stream, use DescribeDeliveryStream. You
        can delete a delivery stream only if it is in one of the following
        states: ``ACTIVE``, ``DELETING``, ``CREATING_FAILED``, or
        ``DELETING_FAILED``. You can't delete a delivery stream that is in the
        ``CREATING`` state. While the deletion request is in process, the
        delivery stream is in the ``DELETING`` state.

        While the delivery stream is in the ``DELETING`` state, the service
        might continue to accept records, but it doesn't make any guarantees
        with respect to delivering the data. Therefore, as a best practice,
        first stop any applications that are sending records before you delete a
        delivery stream.

        :param delivery_stream_name: The name of the delivery stream.
        :param allow_force_delete: Set this to true if you want to delete the delivery stream even if
        Kinesis Data Firehose is unable to retire the grant for the CMK.
        :returns: DeleteDeliveryStreamOutput
        :raises ResourceInUseException:
        :raises ResourceNotFoundException:
        """
        raise NotImplementedError

    @handler("DescribeDeliveryStream")
    def describe_delivery_stream(
        self,
        context: RequestContext,
        delivery_stream_name: DeliveryStreamName,
        limit: DescribeDeliveryStreamInputLimit = None,
        exclusive_start_destination_id: DestinationId = None,
    ) -> DescribeDeliveryStreamOutput:
        """Describes the specified delivery stream and its status. For example,
        after your delivery stream is created, call ``DescribeDeliveryStream``
        to see whether the delivery stream is ``ACTIVE`` and therefore ready for
        data to be sent to it.

        If the status of a delivery stream is ``CREATING_FAILED``, this status
        doesn't change, and you can't invoke CreateDeliveryStream again on it.
        However, you can invoke the DeleteDeliveryStream operation to delete it.
        If the status is ``DELETING_FAILED``, you can force deletion by invoking
        DeleteDeliveryStream again but with
        DeleteDeliveryStreamInput$AllowForceDelete set to true.

        :param delivery_stream_name: The name of the delivery stream.
        :param limit: The limit on the number of destinations to return.
        :param exclusive_start_destination_id: The ID of the destination to start returning the destination
        information.
        :returns: DescribeDeliveryStreamOutput
        :raises ResourceNotFoundException:
        """
        raise NotImplementedError

    @handler("ListDeliveryStreams")
    def list_delivery_streams(
        self,
        context: RequestContext,
        limit: ListDeliveryStreamsInputLimit = None,
        delivery_stream_type: DeliveryStreamType = None,
        exclusive_start_delivery_stream_name: DeliveryStreamName = None,
    ) -> ListDeliveryStreamsOutput:
        """Lists your delivery streams in alphabetical order of their names.

        The number of delivery streams might be too large to return using a
        single call to ``ListDeliveryStreams``. You can limit the number of
        delivery streams returned, using the ``Limit`` parameter. To determine
        whether there are more delivery streams to list, check the value of
        ``HasMoreDeliveryStreams`` in the output. If there are more delivery
        streams to list, you can request them by calling this operation again
        and setting the ``ExclusiveStartDeliveryStreamName`` parameter to the
        name of the last delivery stream returned in the last call.

        :param limit: The maximum number of delivery streams to list.
        :param delivery_stream_type: The delivery stream type.
        :param exclusive_start_delivery_stream_name: The list of delivery streams returned by this call to
        ``ListDeliveryStreams`` will start with the delivery stream whose name
        comes alphabetically immediately after the name you specify in
        ``ExclusiveStartDeliveryStreamName``.
        :returns: ListDeliveryStreamsOutput
        """
        raise NotImplementedError

    @handler("ListTagsForDeliveryStream")
    def list_tags_for_delivery_stream(
        self,
        context: RequestContext,
        delivery_stream_name: DeliveryStreamName,
        exclusive_start_tag_key: TagKey = None,
        limit: ListTagsForDeliveryStreamInputLimit = None,
    ) -> ListTagsForDeliveryStreamOutput:
        """Lists the tags for the specified delivery stream. This operation has a
        limit of five transactions per second per account.

        :param delivery_stream_name: The name of the delivery stream whose tags you want to list.
        :param exclusive_start_tag_key: The key to use as the starting point for the list of tags.
        :param limit: The number of tags to return.
        :returns: ListTagsForDeliveryStreamOutput
        :raises ResourceNotFoundException:
        :raises InvalidArgumentException:
        :raises LimitExceededException:
        """
        raise NotImplementedError

    @handler("PutRecord")
    def put_record(
        self,
        context: RequestContext,
        delivery_stream_name: DeliveryStreamName,
        record: Record,
    ) -> PutRecordOutput:
        """Writes a single data record into an Amazon Kinesis Data Firehose
        delivery stream. To write multiple data records into a delivery stream,
        use PutRecordBatch. Applications using these operations are referred to
        as producers.

        By default, each delivery stream can take in up to 2,000 transactions
        per second, 5,000 records per second, or 5 MB per second. If you use
        PutRecord and PutRecordBatch, the limits are an aggregate across these
        two operations for each delivery stream. For more information about
        limits and how to request an increase, see `Amazon Kinesis Data Firehose
        Limits <https://docs.aws.amazon.com/firehose/latest/dev/limits.html>`__.

        You must specify the name of the delivery stream and the data record
        when using PutRecord. The data record consists of a data blob that can
        be up to 1,000 KiB in size, and any kind of data. For example, it can be
        a segment from a log file, geographic location data, website clickstream
        data, and so on.

        Kinesis Data Firehose buffers records before delivering them to the
        destination. To disambiguate the data blobs at the destination, a common
        solution is to use delimiters in the data, such as a newline (``\n``) or
        some other character unique within the data. This allows the consumer
        application to parse individual data items when reading the data from
        the destination.

        The ``PutRecord`` operation returns a ``RecordId``, which is a unique
        string assigned to each record. Producer applications can use this ID
        for purposes such as auditability and investigation.

        If the ``PutRecord`` operation throws a ``ServiceUnavailableException``,
        back off and retry. If the exception persists, it is possible that the
        throughput limits have been exceeded for the delivery stream.

        Data records sent to Kinesis Data Firehose are stored for 24 hours from
        the time they are added to a delivery stream as it tries to send the
        records to the destination. If the destination is unreachable for more
        than 24 hours, the data is no longer available.

        Don't concatenate two or more base64 strings to form the data fields of
        your records. Instead, concatenate the raw data, then perform base64
        encoding.

        :param delivery_stream_name: The name of the delivery stream.
        :param record: The record.
        :returns: PutRecordOutput
        :raises ResourceNotFoundException:
        :raises InvalidArgumentException:
        :raises InvalidKMSResourceException:
        :raises ServiceUnavailableException:
        """
        raise NotImplementedError

    @handler("PutRecordBatch")
    def put_record_batch(
        self,
        context: RequestContext,
        delivery_stream_name: DeliveryStreamName,
        records: PutRecordBatchRequestEntryList,
    ) -> PutRecordBatchOutput:
        """Writes multiple data records into a delivery stream in a single call,
        which can achieve higher throughput per producer than when writing
        single records. To write single data records into a delivery stream, use
        PutRecord. Applications using these operations are referred to as
        producers.

        For information about service quota, see `Amazon Kinesis Data Firehose
        Quota <https://docs.aws.amazon.com/firehose/latest/dev/limits.html>`__.

        Each PutRecordBatch request supports up to 500 records. Each record in
        the request can be as large as 1,000 KB (before base64 encoding), up to
        a limit of 4 MB for the entire request. These limits cannot be changed.

        You must specify the name of the delivery stream and the data record
        when using PutRecord. The data record consists of a data blob that can
        be up to 1,000 KB in size, and any kind of data. For example, it could
        be a segment from a log file, geographic location data, website
        clickstream data, and so on.

        Kinesis Data Firehose buffers records before delivering them to the
        destination. To disambiguate the data blobs at the destination, a common
        solution is to use delimiters in the data, such as a newline (``\n``) or
        some other character unique within the data. This allows the consumer
        application to parse individual data items when reading the data from
        the destination.

        The PutRecordBatch response includes a count of failed records,
        ``FailedPutCount``, and an array of responses, ``RequestResponses``.
        Even if the PutRecordBatch call succeeds, the value of
        ``FailedPutCount`` may be greater than 0, indicating that there are
        records for which the operation didn't succeed. Each entry in the
        ``RequestResponses`` array provides additional information about the
        processed record. It directly correlates with a record in the request
        array using the same ordering, from the top to the bottom. The response
        array always includes the same number of records as the request array.
        ``RequestResponses`` includes both successfully and unsuccessfully
        processed records. Kinesis Data Firehose tries to process all records in
        each PutRecordBatch request. A single record failure does not stop the
        processing of subsequent records.

        A successfully processed record includes a ``RecordId`` value, which is
        unique for the record. An unsuccessfully processed record includes
        ``ErrorCode`` and ``ErrorMessage`` values. ``ErrorCode`` reflects the
        type of error, and is one of the following values:
        ``ServiceUnavailableException`` or ``InternalFailure``. ``ErrorMessage``
        provides more detailed information about the error.

        If there is an internal server error or a timeout, the write might have
        completed or it might have failed. If ``FailedPutCount`` is greater than
        0, retry the request, resending only those records that might have
        failed processing. This minimizes the possible duplicate records and
        also reduces the total bytes sent (and corresponding charges). We
        recommend that you handle any duplicates at the destination.

        If PutRecordBatch throws ``ServiceUnavailableException``, back off and
        retry. If the exception persists, it is possible that the throughput
        limits have been exceeded for the delivery stream.

        Data records sent to Kinesis Data Firehose are stored for 24 hours from
        the time they are added to a delivery stream as it attempts to send the
        records to the destination. If the destination is unreachable for more
        than 24 hours, the data is no longer available.

        Don't concatenate two or more base64 strings to form the data fields of
        your records. Instead, concatenate the raw data, then perform base64
        encoding.

        :param delivery_stream_name: The name of the delivery stream.
        :param records: One or more records.
        :returns: PutRecordBatchOutput
        :raises ResourceNotFoundException:
        :raises InvalidArgumentException:
        :raises InvalidKMSResourceException:
        :raises ServiceUnavailableException:
        """
        raise NotImplementedError

    @handler("StartDeliveryStreamEncryption")
    def start_delivery_stream_encryption(
        self,
        context: RequestContext,
        delivery_stream_name: DeliveryStreamName,
        delivery_stream_encryption_configuration_input: DeliveryStreamEncryptionConfigurationInput = None,
    ) -> StartDeliveryStreamEncryptionOutput:
        """Enables server-side encryption (SSE) for the delivery stream.

        This operation is asynchronous. It returns immediately. When you invoke
        it, Kinesis Data Firehose first sets the encryption status of the stream
        to ``ENABLING``, and then to ``ENABLED``. The encryption status of a
        delivery stream is the ``Status`` property in
        DeliveryStreamEncryptionConfiguration. If the operation fails, the
        encryption status changes to ``ENABLING_FAILED``. You can continue to
        read and write data to your delivery stream while the encryption status
        is ``ENABLING``, but the data is not encrypted. It can take up to 5
        seconds after the encryption status changes to ``ENABLED`` before all
        records written to the delivery stream are encrypted. To find out
        whether a record or a batch of records was encrypted, check the response
        elements PutRecordOutput$Encrypted and PutRecordBatchOutput$Encrypted,
        respectively.

        To check the encryption status of a delivery stream, use
        DescribeDeliveryStream.

        Even if encryption is currently enabled for a delivery stream, you can
        still invoke this operation on it to change the ARN of the CMK or both
        its type and ARN. If you invoke this method to change the CMK, and the
        old CMK is of type ``CUSTOMER_MANAGED_CMK``, Kinesis Data Firehose
        schedules the grant it had on the old CMK for retirement. If the new CMK
        is of type ``CUSTOMER_MANAGED_CMK``, Kinesis Data Firehose creates a
        grant that enables it to use the new CMK to encrypt and decrypt data and
        to manage the grant.

        If a delivery stream already has encryption enabled and then you invoke
        this operation to change the ARN of the CMK or both its type and ARN and
        you get ``ENABLING_FAILED``, this only means that the attempt to change
        the CMK failed. In this case, encryption remains enabled with the old
        CMK.

        If the encryption status of your delivery stream is ``ENABLING_FAILED``,
        you can invoke this operation again with a valid CMK. The CMK must be
        enabled and the key policy mustn't explicitly deny the permission for
        Kinesis Data Firehose to invoke KMS encrypt and decrypt operations.

        You can enable SSE for a delivery stream only if it's a delivery stream
        that uses ``DirectPut`` as its source.

        The ``StartDeliveryStreamEncryption`` and
        ``StopDeliveryStreamEncryption`` operations have a combined limit of 25
        calls per delivery stream per 24 hours. For example, you reach the limit
        if you call ``StartDeliveryStreamEncryption`` 13 times and
        ``StopDeliveryStreamEncryption`` 12 times for the same delivery stream
        in a 24-hour period.

        :param delivery_stream_name: The name of the delivery stream for which you want to enable server-side
        encryption (SSE).
        :param delivery_stream_encryption_configuration_input: Used to specify the type and Amazon Resource Name (ARN) of the KMS key
        needed for Server-Side Encryption (SSE).
        :returns: StartDeliveryStreamEncryptionOutput
        :raises ResourceNotFoundException:
        :raises ResourceInUseException:
        :raises InvalidArgumentException:
        :raises LimitExceededException:
        :raises InvalidKMSResourceException:
        """
        raise NotImplementedError

    @handler("StopDeliveryStreamEncryption")
    def stop_delivery_stream_encryption(
        self, context: RequestContext, delivery_stream_name: DeliveryStreamName
    ) -> StopDeliveryStreamEncryptionOutput:
        """Disables server-side encryption (SSE) for the delivery stream.

        This operation is asynchronous. It returns immediately. When you invoke
        it, Kinesis Data Firehose first sets the encryption status of the stream
        to ``DISABLING``, and then to ``DISABLED``. You can continue to read and
        write data to your stream while its status is ``DISABLING``. It can take
        up to 5 seconds after the encryption status changes to ``DISABLED``
        before all records written to the delivery stream are no longer subject
        to encryption. To find out whether a record or a batch of records was
        encrypted, check the response elements PutRecordOutput$Encrypted and
        PutRecordBatchOutput$Encrypted, respectively.

        To check the encryption state of a delivery stream, use
        DescribeDeliveryStream.

        If SSE is enabled using a customer managed CMK and then you invoke
        ``StopDeliveryStreamEncryption``, Kinesis Data Firehose schedules the
        related KMS grant for retirement and then retires it after it ensures
        that it is finished delivering records to the destination.

        The ``StartDeliveryStreamEncryption`` and
        ``StopDeliveryStreamEncryption`` operations have a combined limit of 25
        calls per delivery stream per 24 hours. For example, you reach the limit
        if you call ``StartDeliveryStreamEncryption`` 13 times and
        ``StopDeliveryStreamEncryption`` 12 times for the same delivery stream
        in a 24-hour period.

        :param delivery_stream_name: The name of the delivery stream for which you want to disable
        server-side encryption (SSE).
        :returns: StopDeliveryStreamEncryptionOutput
        :raises ResourceNotFoundException:
        :raises ResourceInUseException:
        :raises InvalidArgumentException:
        :raises LimitExceededException:
        """
        raise NotImplementedError

    @handler("TagDeliveryStream")
    def tag_delivery_stream(
        self,
        context: RequestContext,
        delivery_stream_name: DeliveryStreamName,
        tags: TagDeliveryStreamInputTagList,
    ) -> TagDeliveryStreamOutput:
        """Adds or updates tags for the specified delivery stream. A tag is a
        key-value pair that you can define and assign to AWS resources. If you
        specify a tag that already exists, the tag value is replaced with the
        value that you specify in the request. Tags are metadata. For example,
        you can add friendly names and descriptions or other types of
        information that can help you distinguish the delivery stream. For more
        information about tags, see `Using Cost Allocation
        Tags <https://docs.aws.amazon.com/awsaccountbilling/latest/aboutv2/cost-alloc-tags.html>`__
        in the *AWS Billing and Cost Management User Guide*.

        Each delivery stream can have up to 50 tags.

        This operation has a limit of five transactions per second per account.

        :param delivery_stream_name: The name of the delivery stream to which you want to add the tags.
        :param tags: A set of key-value pairs to use to create the tags.
        :returns: TagDeliveryStreamOutput
        :raises ResourceNotFoundException:
        :raises ResourceInUseException:
        :raises InvalidArgumentException:
        :raises LimitExceededException:
        """
        raise NotImplementedError

    @handler("UntagDeliveryStream")
    def untag_delivery_stream(
        self,
        context: RequestContext,
        delivery_stream_name: DeliveryStreamName,
        tag_keys: TagKeyList,
    ) -> UntagDeliveryStreamOutput:
        """Removes tags from the specified delivery stream. Removed tags are
        deleted, and you can't recover them after this operation successfully
        completes.

        If you specify a tag that doesn't exist, the operation ignores it.

        This operation has a limit of five transactions per second per account.

        :param delivery_stream_name: The name of the delivery stream.
        :param tag_keys: A list of tag keys.
        :returns: UntagDeliveryStreamOutput
        :raises ResourceNotFoundException:
        :raises ResourceInUseException:
        :raises InvalidArgumentException:
        :raises LimitExceededException:
        """
        raise NotImplementedError

    @handler("UpdateDestination")
    def update_destination(
        self,
        context: RequestContext,
        delivery_stream_name: DeliveryStreamName,
        current_delivery_stream_version_id: DeliveryStreamVersionId,
        destination_id: DestinationId,
        s3_destination_update: S3DestinationUpdate = None,
        extended_s3_destination_update: ExtendedS3DestinationUpdate = None,
        redshift_destination_update: RedshiftDestinationUpdate = None,
        elasticsearch_destination_update: ElasticsearchDestinationUpdate = None,
        amazonopensearchservice_destination_update: AmazonopensearchserviceDestinationUpdate = None,
        splunk_destination_update: SplunkDestinationUpdate = None,
        http_endpoint_destination_update: HttpEndpointDestinationUpdate = None,
    ) -> UpdateDestinationOutput:
        """Updates the specified destination of the specified delivery stream.

        Use this operation to change the destination type (for example, to
        replace the Amazon S3 destination with Amazon Redshift) or change the
        parameters associated with a destination (for example, to change the
        bucket name of the Amazon S3 destination). The update might not occur
        immediately. The target delivery stream remains active while the
        configurations are updated, so data writes to the delivery stream can
        continue during this process. The updated configurations are usually
        effective within a few minutes.

        Switching between Amazon ES and other services is not supported. For an
        Amazon ES destination, you can only update to another Amazon ES
        destination.

        If the destination type is the same, Kinesis Data Firehose merges the
        configuration parameters specified with the destination configuration
        that already exists on the delivery stream. If any of the parameters are
        not specified in the call, the existing values are retained. For
        example, in the Amazon S3 destination, if EncryptionConfiguration is not
        specified, then the existing ``EncryptionConfiguration`` is maintained
        on the destination.

        If the destination type is not the same, for example, changing the
        destination from Amazon S3 to Amazon Redshift, Kinesis Data Firehose
        does not merge any parameters. In this case, all parameters must be
        specified.

        Kinesis Data Firehose uses ``CurrentDeliveryStreamVersionId`` to avoid
        race conditions and conflicting merges. This is a required field, and
        the service updates the configuration only if the existing configuration
        has a version ID that matches. After the update is applied successfully,
        the version ID is updated, and can be retrieved using
        DescribeDeliveryStream. Use the new version ID to set
        ``CurrentDeliveryStreamVersionId`` in the next call.

        :param delivery_stream_name: The name of the delivery stream.
        :param current_delivery_stream_version_id: Obtain this value from the ``VersionId`` result of
        DeliveryStreamDescription.
        :param destination_id: The ID of the destination.
        :param s3_destination_update: [Deprecated] Describes an update for a destination in Amazon S3.
        :param extended_s3_destination_update: Describes an update for a destination in Amazon S3.
        :param redshift_destination_update: Describes an update for a destination in Amazon Redshift.
        :param elasticsearch_destination_update: Describes an update for a destination in Amazon ES.
        :param amazonopensearchservice_destination_update: .
        :param splunk_destination_update: Describes an update for a destination in Splunk.
        :param http_endpoint_destination_update: Describes an update to the specified HTTP endpoint destination.
        :returns: UpdateDestinationOutput
        :raises InvalidArgumentException:
        :raises ResourceInUseException:
        :raises ResourceNotFoundException:
        :raises ConcurrentModificationException:
        """
        raise NotImplementedError



import json
import logging
import re
from typing import Dict, Optional, Tuple
from urllib.parse import parse_qs, urlparse

from moto.apigateway import models as apigateway_models
from moto.apigateway.exceptions import NoIntegrationDefined, UsagePlanNotFoundException
from moto.apigateway.responses import APIGatewayResponse
from moto.core.utils import camelcase_to_underscores

from localstack import config
from localstack.constants import TEST_AWS_ACCOUNT_ID
from localstack.services.apigateway.helpers import (
    TAG_KEY_CUSTOM_ID,
    apply_json_patch_safe,
    import_api_from_openapi_spec,
)
from localstack.services.infra import start_moto_server
from localstack.utils.common import DelSafeDict, short_uid, str_to_bool, to_str

LOG = logging.getLogger(__name__)

# additional REST API attributes
REST_API_ATTRIBUTES = [
    "apiKeySource",
    "disableExecuteApiEndpoint",
    "minimumCompressionSize",
]


def apply_patches():
    def apigateway_models_Stage_init(
        self, cacheClusterEnabled=False, cacheClusterSize=None, **kwargs
    ):
        apigateway_models_Stage_init_orig(
            self,
            cacheClusterEnabled=cacheClusterEnabled,
            cacheClusterSize=cacheClusterSize,
            **kwargs,
        )

        if (cacheClusterSize or cacheClusterEnabled) and not self.get("cacheClusterStatus"):
            self["cacheClusterStatus"] = "AVAILABLE"

    apigateway_models_Stage_init_orig = apigateway_models.Stage.__init__
    apigateway_models.Stage.__init__ = apigateway_models_Stage_init

    def apigateway_models_backend_delete_method(self, function_id, resource_id, method_type):
        resource = self.get_resource(function_id, resource_id)
        method = resource.get_method(method_type)
        if not method:
            return
        return resource.resource_methods.pop(method_type)

    def apigateway_models_resource_delete_integration(self, method_type):
        if method_type in self.resource_methods:
            return self.resource_methods[method_type].pop("methodIntegration", {})

        return {}

    def apigateway_models_Integration_init(
        self,
        integration_type,
        uri,
        http_method,
        request_templates=None,
        pass_through_behavior="WHEN_NO_MATCH",
        cache_key_parameters=None,
        *args,
        **kwargs,
    ):
        if cache_key_parameters is None:
            cache_key_parameters = []
        apigateway_models_Integration_init_orig(
            self,
            integration_type=integration_type,
            uri=uri,
            http_method=http_method,
            request_templates=request_templates,
            *args,
            **kwargs,
        )

        self["passthroughBehavior"] = pass_through_behavior
        self["cacheKeyParameters"] = cache_key_parameters
        self["cacheNamespace"] = self.get("cacheNamespace") or short_uid()

        # httpMethod not present in response if integration_type is None, verified against AWS
        if integration_type == "MOCK":
            self["httpMethod"] = None
        if request_templates:
            self["requestTemplates"] = request_templates

    def apigateway_models_backend_put_rest_api(self, function_id, body, query_params):
        rest_api = self.get_rest_api(function_id)
        return import_api_from_openapi_spec(rest_api, function_id, body, query_params)

    def _patch_api_gateway_entity(self, entity: Dict) -> Optional[Tuple[int, Dict, str]]:
        not_supported_attributes = ["/id", "/region_name", "/create_date"]

        patch_operations = self._get_param("patchOperations")

        model_attributes = list(entity.keys())
        for operation in patch_operations:
            if operation["path"].strip("/") in REST_API_ATTRIBUTES:
                operation["path"] = camelcase_to_underscores(operation["path"])
            path_start = operation["path"].strip("/").split("/")[0]
            path_start_usc = camelcase_to_underscores(path_start)
            if path_start not in model_attributes and path_start_usc in model_attributes:
                operation["path"] = operation["path"].replace(path_start, path_start_usc)
            if operation["path"] in not_supported_attributes:
                msg = "Invalid patch path %s" % (operation["path"])
                return 400, {}, msg

        apply_json_patch_safe(entity, patch_operations, in_place=True)

    def apigateway_response_restapis_individual(self, request, full_url, headers):
        if request.method in ["GET", "DELETE"]:
            return apigateway_response_restapis_individual_orig(self, request, full_url, headers)

        self.setup_class(request, full_url, headers)
        function_id = self.path.replace("/restapis/", "", 1).split("/")[0]

        if self.method == "PATCH":
            rest_api = self.backend.apis.get(function_id)
            if not rest_api:
                msg = "Invalid API identifier specified %s:%s" % (
                    TEST_AWS_ACCOUNT_ID,
                    function_id,
                )
                return 404, {}, msg

            if not isinstance(rest_api.__dict__, DelSafeDict):
                rest_api.__dict__ = DelSafeDict(rest_api.__dict__)

            result = _patch_api_gateway_entity(self, rest_api.__dict__)
            if result is not None:
                return result

            # fix data types after patches have been applied
            rest_api.minimum_compression_size = int(rest_api.minimum_compression_size or -1)
            endpoint_configs = rest_api.endpoint_configuration or {}
            if isinstance(endpoint_configs.get("vpcEndpointIds"), str):
                endpoint_configs["vpcEndpointIds"] = [endpoint_configs["vpcEndpointIds"]]

            return 200, {}, json.dumps(self.backend.get_rest_api(function_id).to_dict())

        # handle import rest_api via swagger file
        if self.method == "PUT":
            body = json.loads(to_str(self.body))
            rest_api = self.backend.put_rest_api(function_id, body, self.querystring)
            return 200, {}, json.dumps(rest_api.to_dict())

        return 400, {}, ""

    def apigateway_response_resource_individual(self, request, full_url, headers):
        if request.method in ["GET", "POST", "DELETE"]:
            return apigateway_response_resource_individual_orig(self, request, full_url, headers)

        self.setup_class(request, full_url, headers)
        function_id = self.path.replace("/restapis/", "", 1).split("/")[0]

        if self.method == "PATCH":
            resource_id = self.path.split("/")[4]
            resource = self.backend.get_resource(function_id, resource_id)
            if not isinstance(resource.__dict__, DelSafeDict):
                resource.__dict__ = DelSafeDict(resource.__dict__)
            result = _patch_api_gateway_entity(self, resource.__dict__)
            if result is not None:
                return result
            return 200, {}, json.dumps(resource.to_dict())

        return 404, {}, ""

    def apigateway_response_resource_methods(self, request, *args, **kwargs):
        result = apigateway_response_resource_methods_orig(self, request, *args, **kwargs)

        if self.method == "PUT" and self._get_param("requestParameters"):
            request_parameters = self._get_param("requestParameters")
            url_path_parts = self.path.split("/")
            function_id = url_path_parts[2]
            resource_id = url_path_parts[4]
            method_type = url_path_parts[6]
            resource = self.backend.get_resource(function_id, resource_id)
            resource.resource_methods[method_type]["requestParameters"] = request_parameters
            method = resource.resource_methods[method_type]
            result = 200, {}, json.dumps(method)
        if len(result) != 3:
            return result
        authorization_type = self._get_param("authorizationType")
        if authorization_type in ["CUSTOM", "COGNITO_USER_POOLS"]:
            data = json.loads(result[2])
            if not data.get("authorizerId"):
                payload = json.loads(to_str(request.data))
                if "authorizerId" in payload:
                    data["authorizerId"] = payload["authorizerId"]
                    result = result[0], result[1], json.dumps(data)
        return result

    def apigateway_response_integrations(self, request, *args, **kwargs):
        result = apigateway_response_integrations_orig(self, request, *args, **kwargs)

        if self.method not in ["PUT", "PATCH"]:
            return result

        url_path_parts = self.path.split("/")
        function_id = url_path_parts[2]
        resource_id = url_path_parts[4]
        method_type = url_path_parts[6]

        integration = self.backend.get_integration(function_id, resource_id, method_type)
        if not integration:
            return result

        if self.method == "PUT":
            timeout_milliseconds = self._get_param("timeoutInMillis")
            request_parameters = self._get_param("requestParameters") or {}
            cache_key_parameters = self._get_param("cacheKeyParameters") or []
            content_handling = self._get_param("contentHandling")
            integration["timeoutInMillis"] = timeout_milliseconds
            integration["requestParameters"] = request_parameters
            integration["cacheKeyParameters"] = cache_key_parameters
            integration["contentHandling"] = content_handling
            return 200, {}, json.dumps(integration)

        if self.method == "PATCH":
            patch_operations = self._get_param("patchOperations")
            apply_json_patch_safe(integration, patch_operations, in_place=True)
            # fix data types
            if integration.get("timeoutInMillis"):
                integration["timeoutInMillis"] = int(integration.get("timeoutInMillis"))
            skip_verification = (integration.get("tlsConfig") or {}).get("insecureSkipVerification")
            if skip_verification:
                integration["tlsConfig"]["insecureSkipVerification"] = str_to_bool(
                    skip_verification
                )

        return result

    def apigateway_response_integration_responses(self, request, *args, **kwargs):
        result = apigateway_response_integration_responses_orig(self, request, *args, **kwargs)
        response_parameters = self._get_param("responseParameters")

        if self.method == "PUT" and response_parameters:
            url_path_parts = self.path.split("/")
            function_id = url_path_parts[2]
            resource_id = url_path_parts[4]
            method_type = url_path_parts[6]
            status_code = url_path_parts[9]

            integration_response = self.backend.get_integration_response(
                function_id, resource_id, method_type, status_code
            )
            integration_response["responseParameters"] = response_parameters

            return 200, {}, json.dumps(integration_response)

        return result

    def apigateway_response_resource_method_responses(self, request, *args, **kwargs):
        result = apigateway_response_resource_method_responses_orig(self, request, *args, **kwargs)
        response_parameters = self._get_param("responseParameters")

        if self.method == "PUT" and response_parameters:
            url_path_parts = self.path.split("/")
            function_id = url_path_parts[2]
            resource_id = url_path_parts[4]
            method_type = url_path_parts[6]
            response_code = url_path_parts[8]

            method_response = self.backend.get_method_response(
                function_id, resource_id, method_type, response_code
            )

            method_response["responseParameters"] = response_parameters

            return 200, {}, json.dumps(method_response)

        return result

    def apigateway_response_usage_plan_individual(
        self, request, full_url, headers, *args, **kwargs
    ):
        self.setup_class(request, full_url, headers)
        if self.method == "PATCH":
            url_path_parts = self.path.split("/")
            usage_plan_id = url_path_parts[2]
            patch_operations = self._get_param("patchOperations")
            usage_plan = self.backend.usage_plans.get(usage_plan_id)
            if not usage_plan:
                raise UsagePlanNotFoundException()

            apply_json_patch_safe(usage_plan, patch_operations, in_place=True)
            # fix certain attributes after running the patch updates
            if isinstance(usage_plan.get("apiStages"), (dict, str)):
                usage_plan["apiStages"] = [usage_plan["apiStages"]]
            api_stages = usage_plan.get("apiStages") or []
            for i in range(len(api_stages)):
                if isinstance(api_stages[i], str) and ":" in api_stages[i]:
                    api_id, stage = api_stages[i].split(":")
                    api_stages[i] = {"apiId": api_id, "stage": stage}

            return 200, {}, json.dumps(usage_plan)
        result = apigateway_response_usage_plan_individual_orig(
            self, request, full_url, headers, *args, **kwargs
        )
        return result

    def backend_update_deployment(self, function_id, deployment_id, patch_operations):
        rest_api = self.get_rest_api(function_id)
        deployment = rest_api.get_deployment(deployment_id)
        deployment = deployment or {}
        apply_json_patch_safe(deployment, patch_operations, in_place=True)
        return deployment

    # define json-patch operations for backend models

    def backend_model_apply_operations(self, patch_operations):
        # run pre-actions
        if isinstance(self, apigateway_models.Stage):
            if [op for op in patch_operations if "/accessLogSettings" in op.get("path", "")]:
                self["accessLogSettings"] = self.get("accessLogSettings") or {}
        # apply patches
        apply_json_patch_safe(self, patch_operations, in_place=True)
        # run post-actions
        if isinstance(self, apigateway_models.Stage):
            bool_params = ["cacheClusterEnabled", "tracingEnabled"]
            for bool_param in bool_params:
                if self.get(bool_param):
                    self[bool_param] = str_to_bool(self.get(bool_param))
        return self

    model_classes = [
        apigateway_models.Authorizer,
        apigateway_models.DomainName,
        apigateway_models.Method,
        apigateway_models.MethodResponse,
        apigateway_models.Stage,
    ]
    for model_class in model_classes:
        model_class.apply_operations = (
            model_class.apply_patch_operations
        ) = backend_model_apply_operations

    # fix data types for some json-patch operation values

    def method_apply_operations(self, patch_operations):
        result = method_apply_operations_orig(self, patch_operations)
        params = self.get("requestParameters") or {}
        bool_params_prefixes = ["method.request.querystring", "method.request.header"]
        list_params = ["authorizationScopes"]
        for param, value in params.items():
            for param_prefix in bool_params_prefixes:
                if param.startswith(param_prefix):
                    params[param] = str_to_bool(value)
        for list_param in list_params:
            value = self.get(list_param)
            if value and not isinstance(value, list):
                self[list_param] = [value]
        return result

    method_apply_operations_orig = apigateway_models.Method.apply_operations
    apigateway_models.Method.apply_operations = method_apply_operations

    def method_response_apply_operations(self, patch_operations):
        result = method_response_apply_operations_orig(self, patch_operations)
        params = self.get("responseParameters") or {}
        bool_params_prefixes = ["method.response.querystring", "method.response.header"]
        for param, value in params.items():
            for param_prefix in bool_params_prefixes:
                if param.startswith(param_prefix) and not isinstance(value, bool):
                    params[param] = str(value) in ["true", "True"]
        return result

    method_response_apply_operations_orig = apigateway_models.MethodResponse.apply_operations
    apigateway_models.MethodResponse.apply_operations = method_response_apply_operations

    def stage_apply_operations(self, patch_operations):
        result = stage_apply_operations_orig(self, patch_operations)
        key_mappings = {
            "metrics/enabled": ("metricsEnabled", bool),
            "logging/loglevel": ("loggingLevel", str),
            "logging/dataTrace": ("dataTraceEnabled", bool),
            "throttling/burstLimit": ("throttlingBurstLimit", int),
            "throttling/rateLimit": ("throttlingRateLimit", float),
            "caching/enabled": ("cachingEnabled", bool),
            "caching/ttlInSeconds": ("cacheTtlInSeconds", int),
            "caching/dataEncrypted": ("cacheDataEncrypted", bool),
            "caching/requireAuthorizationForCacheControl": (
                "requireAuthorizationForCacheControl",
                bool,
            ),
            "caching/unauthorizedCacheControlHeaderStrategy": (
                "unauthorizedCacheControlHeaderStrategy",
                str,
            ),
        }

        def cast_value(value, value_type):
            if value is None:
                return value
            if value_type == bool:
                return str(value) in ["true", "True"]
            return value_type(value)

        method_settings = self["methodSettings"] = self.get("methodSettings") or {}
        for operation in patch_operations:
            path = operation["path"]
            parts = path.strip("/").split("/")
            if len(parts) >= 4:
                if operation["op"] not in ["add", "replace"]:
                    continue
                key1 = "/".join(parts[:-2])
                setting_key = "%s/%s" % (parts[-2], parts[-1])
                setting_name, setting_type = key_mappings.get(setting_key)
                keys = [key1]
                for key in keys:
                    setting = method_settings[key] = method_settings.get(key) or {}
                    value = operation.get("value")
                    value = cast_value(value, setting_type)
                    setting[setting_name] = value
            if operation["op"] == "remove":
                method_settings.pop(path, None)
                method_settings.pop(path.lstrip("/"), None)
        return result

    stage_apply_operations_orig = apigateway_models.Stage.apply_operations
    apigateway_models.Stage.apply_operations = stage_apply_operations

    # patch integration error responses

    def apigateway_models_resource_get_integration(self, method_type):
        resource_method = self.resource_methods.get(method_type, {})
        if "methodIntegration" not in resource_method:
            raise NoIntegrationDefined()
        return resource_method["methodIntegration"]

    if not hasattr(apigateway_models.APIGatewayBackend, "put_rest_api"):
        apigateway_response_restapis_individual_orig = APIGatewayResponse.restapis_individual
        APIGatewayResponse.restapis_individual = apigateway_response_restapis_individual
        apigateway_response_resource_individual_orig = APIGatewayResponse.resource_individual
        APIGatewayResponse.resource_individual = apigateway_response_resource_individual
        apigateway_models.APIGatewayBackend.put_rest_api = apigateway_models_backend_put_rest_api

    if not hasattr(apigateway_models.APIGatewayBackend, "delete_method"):
        apigateway_models.APIGatewayBackend.delete_method = apigateway_models_backend_delete_method

    if not hasattr(apigateway_models.APIGatewayBackend, "update_deployment"):
        apigateway_models.APIGatewayBackend.update_deployment = backend_update_deployment

    apigateway_models_RestAPI_to_dict_orig = apigateway_models.RestAPI.to_dict

    def apigateway_models_RestAPI_to_dict(self):
        resp = apigateway_models_RestAPI_to_dict_orig(self)
        resp["policy"] = None
        if self.policy:
            # Strip whitespaces for TF compatibility (not entirely sure why we need double-dumps,
            # but otherwise: "error normalizing policy JSON: invalid character 'V' after top-level value")
            resp["policy"] = json.dumps(json.dumps(json.loads(self.policy), separators=(",", ":")))[
                1:-1
            ]
        for attr in REST_API_ATTRIBUTES:
            if attr not in resp:
                resp[attr] = getattr(self, camelcase_to_underscores(attr), None)
        resp["disableExecuteApiEndpoint"] = bool(
            re.match(
                r"true",
                resp.get("disableExecuteApiEndpoint") or "",
                flags=re.IGNORECASE,
            )
        )

        return resp

    apigateway_response_restapis_orig = APIGatewayResponse.restapis

    # https://github.com/localstack/localstack/issues/171
    def apigateway_response_restapis(self, request, full_url, headers):
        parsed_qs = parse_qs(urlparse(full_url).query)
        modes = parsed_qs.get("mode", [])

        status, _, rest_api = apigateway_response_restapis_orig(self, request, full_url, headers)

        if "import" not in modes:
            return status, _, rest_api

        function_id = json.loads(rest_api)["id"]
        body = json.loads(request.data.decode("utf-8"))
        self.backend.put_rest_api(function_id, body, parsed_qs)

        return 200, {}, rest_api

    def individual_deployment(self, request, full_url, headers, *args, **kwargs):
        result = individual_deployment_orig(self, request, full_url, headers, *args, **kwargs)
        if self.method == "PATCH" and len(result) >= 3 and result[2] in ["null", None, str(None)]:
            url_path_parts = self.path.split("/")
            function_id = url_path_parts[2]
            deployment_id = url_path_parts[4]
            patch_operations = self._get_param("patchOperations")
            deployment = self.backend.update_deployment(
                function_id, deployment_id, patch_operations
            )
            return 200, {}, json.dumps(deployment)
        return result

    # patch create_rest_api to allow using static API IDs defined via tags

    def create_rest_api(self, *args, tags={}, **kwargs):
        result = create_rest_api_orig(self, *args, tags=tags, **kwargs)
        tags = tags or {}
        custom_id = tags.get(TAG_KEY_CUSTOM_ID)
        if custom_id:
            self.apis.pop(result.id)
            result.id = custom_id
            self.apis[custom_id] = result
        return result

    create_rest_api_orig = apigateway_models.APIGatewayBackend.create_rest_api
    apigateway_models.APIGatewayBackend.create_rest_api = create_rest_api
    apigateway_models.Resource.get_integration = apigateway_models_resource_get_integration
    apigateway_models.Resource.delete_integration = apigateway_models_resource_delete_integration
    apigateway_response_resource_methods_orig = APIGatewayResponse.resource_methods
    APIGatewayResponse.resource_methods = apigateway_response_resource_methods
    individual_deployment_orig = APIGatewayResponse.individual_deployment
    APIGatewayResponse.individual_deployment = individual_deployment
    apigateway_response_integrations_orig = APIGatewayResponse.integrations
    APIGatewayResponse.integrations = apigateway_response_integrations
    apigateway_response_integration_responses_orig = APIGatewayResponse.integration_responses
    APIGatewayResponse.integration_responses = apigateway_response_integration_responses
    apigateway_response_resource_method_responses_orig = (
        APIGatewayResponse.resource_method_responses
    )
    APIGatewayResponse.resource_method_responses = apigateway_response_resource_method_responses
    apigateway_response_usage_plan_individual_orig = APIGatewayResponse.usage_plan_individual
    APIGatewayResponse.usage_plan_individual = apigateway_response_usage_plan_individual
    apigateway_models_Integration_init_orig = apigateway_models.Integration.__init__
    apigateway_models.Integration.__init__ = apigateway_models_Integration_init
    apigateway_models.RestAPI.to_dict = apigateway_models_RestAPI_to_dict
    APIGatewayResponse.restapis = apigateway_response_restapis


def start_apigateway(port=None, backend_port=None, asynchronous=None, update_listener=None):
    port = port or config.service_port("apigateway")
    apply_patches()
    result = start_moto_server(
        key="apigateway",
        name="API Gateway",
        asynchronous=asynchronous,
        port=port,
        backend_port=backend_port,
        update_listener=update_listener,
    )
    return result
